def test_target(target):
    barebox = target.get_driver("CommandProtocol")
    target.activate(barebox)

    stdout, stderr, returncode = barebox.run("version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "barebox" in "\n".join(stdout)

    stdout, stderr, returncode = barebox.run("false")
    assert returncode == 1
    assert not stdout
    assert not stderr



"""The remote.client module contains the functionality to connect to a
coordinator, acquire a place and interact with the connected resources"""

import argparse
import asyncio
import contextlib
from contextvars import ContextVar
import enum
import os
import pathlib
import subprocess
import traceback
import logging
import signal
import sys
import shlex
import shutil
import json
import itertools
import ipaddress
from textwrap import indent
from socket import gethostname
from getpass import getuser
from collections import defaultdict, OrderedDict
from datetime import datetime
from pprint import pformat
from typing import Any, Dict

import attr
import grpc

# TODO: drop if Python >= 3.11 guaranteed
from exceptiongroup import ExceptionGroup  # pylint: disable=redefined-builtin

from .common import (
    ResourceEntry,
    ResourceMatch,
    Place,
    Reservation,
    ReservationState,
    TAG_KEY,
    TAG_VAL,
    queue_as_aiter,
)
from .. import Environment, Target, target_factory
from ..exceptions import NoDriverFoundError, NoResourceFoundError, InvalidConfigError
from .generated import labgrid_coordinator_pb2, labgrid_coordinator_pb2_grpc
from ..resource.remote import RemotePlaceManager, RemotePlace
from ..util import diff_dict, flat_dict, dump, atomic_replace, labgrid_version, Timeout
from ..util.proxy import proxymanager
from ..util.helper import processwrapper
from ..driver import Mode, ExecutionError
from ..logging import basicConfig, StepLogger

# This is a workround for the gRPC issue
# https://github.com/grpc/grpc/issues/38679.
# Since Python 3.12, an empty exception message is printed from gRPC
# during shutdown, although nothing seems to go wrong. As this is
# confusing for users, suppress the message by adding an indirection.
sys.excepthook = lambda type, value, traceback: sys.__excepthook__(type, value, traceback)


class Error(Exception):
    def __str__(self):
        return f"Error: {' '.join(self.args)}"


class UserError(Error):
    pass


class ServerError(Error):
    pass


class InteractiveCommandError(Error):
    def __init__(self: Error, msg: str, exitcode: int):
        super(InteractiveCommandError, self).__init__(msg)
        self.exitcode = exitcode


class ErrorGroup(ExceptionGroup):
    def __str__(self):
        # TODO: drop pylint disable once https://github.com/pylint-dev/pylint/issues/8985 is fixed
        errors_combined = "\n".join(f"- {' '.join(e.args)}" for e in self.exceptions)  # pylint: disable=not-an-iterable
        return f"{self.message}:\n{errors_combined}"


@attr.s(eq=False)
class ClientSession:
    """The ClientSession encapsulates all the actions a Client can invoke on
    the coordinator."""

    address = attr.ib(validator=attr.validators.instance_of(str))
    loop = attr.ib(validator=attr.validators.instance_of(asyncio.BaseEventLoop))
    env = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(Environment)))
    role = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    prog = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    args = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(argparse.Namespace)))
    monitor = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    def gethostname(self):
        return os.environ.get("LG_HOSTNAME", gethostname())

    def getuser(self):
        return os.environ.get("LG_USERNAME", getuser())

    def __attrs_post_init__(self):
        """Actions which are executed if a connection is successfully opened."""
        self.stopping = asyncio.Event()

        # It seems since https://github.com/grpc/grpc/pull/34647, the
        # ping_timeout_ms default of 60 seconds overrides keepalive_timeout_ms,
        # so set it as well.
        # Use GRPC_VERBOSITY=DEBUG GRPC_TRACE=http_keepalive for debugging.
        channel_options = [
            ("grpc.keepalive_time_ms", 7500),  # 7.5 seconds
            ("grpc.keepalive_timeout_ms", 10000),  # 10 seconds
            ("grpc.http2.ping_timeout_ms", 10000),  # 10 seconds
            ("grpc.http2.max_pings_without_data", 0),  # no limit
        ]

        self.channel = grpc.aio.insecure_channel(
            target=self.address,
            options=channel_options,
        )
        self.stub = labgrid_coordinator_pb2_grpc.CoordinatorStub(self.channel)

        self.out_queue = asyncio.Queue()
        self.stream_call = None
        self.pump_task = None
        self.sync_id = itertools.count(start=1)
        self.sync_events = {}

    async def start(self):
        """Starts receiving resource and place updates from the coordinator."""
        self.resources = {}
        self.places = {}

        self.pump_task = self.loop.create_task(self.message_pump())
        msg = labgrid_coordinator_pb2.ClientInMessage()
        msg.startup.version = labgrid_version()
        msg.startup.name = f"{self.gethostname()}/{self.getuser()}"
        self.out_queue.put_nowait(msg)
        msg = labgrid_coordinator_pb2.ClientInMessage()
        msg.subscribe.all_places = True
        self.out_queue.put_nowait(msg)
        msg = labgrid_coordinator_pb2.ClientInMessage()
        msg.subscribe.all_resources = True
        self.out_queue.put_nowait(msg)
        await self.sync_with_coordinator()
        if self.stopping.is_set():
            raise ServerError("Could not connect to coordinator")

    async def stop(self):
        """Stops stream for resource and place updates started with ClientSession.start()."""
        self.out_queue.put_nowait(None)  # let the sender side exit gracefully
        if self.stream_call:
            self.stream_call.cancel()
        try:
            await self.pump_task
        except asyncio.CancelledError:
            pass
        self.cancel_pending_syncs()

    async def close(self):
        """Closes the channel to the coordinator."""
        await self.channel.close()

    async def sync_with_coordinator(self):
        """Wait for coordinator to process all previous messages in stream."""
        identifier = next(self.sync_id)
        event = self.sync_events[identifier] = asyncio.Event()
        msg = labgrid_coordinator_pb2.ClientInMessage()
        msg.sync.id = identifier
        logging.debug("sending sync %s", identifier)
        self.out_queue.put_nowait(msg)
        await event.wait()
        if self.stopping.is_set():
            logging.debug("sync %s failed", identifier)
        else:
            logging.debug("received sync %s", identifier)
        return not self.stopping.is_set()

    def cancel_pending_syncs(self):
        """Cancel all pending ClientSession.sync_with_coordinator() calls."""
        assert self.stopping.is_set()  # only call when something has gone wrong
        while True:
            try:
                identifier, event = self.sync_events.popitem()
                logging.debug("cancelling %s %s", identifier, event)
                event.set()
            except KeyError:
                break

    async def message_pump(self):
        """Task for receiving resource and place updates."""
        got_message = False
        try:
            self.stream_call = call = self.stub.ClientStream(queue_as_aiter(self.out_queue))
            async for out_msg in call:
                out_msg: labgrid_coordinator_pb2.ClientOutMessage
                got_message = True
                logging.debug("out_msg from coordinator: %s", out_msg)
                for update in out_msg.updates:
                    update_kind = update.WhichOneof("kind")
                    if update_kind == "resource":
                        resource: labgrid_coordinator_pb2.Resource = update.resource
                        await self.on_resource_changed(
                            resource.path.exporter_name,
                            resource.path.group_name,
                            resource.path.resource_name,
                            ResourceEntry.data_from_pb2(resource),
                        )
                    elif update_kind == "del_resource":
                        resource_path: labgrid_coordinator_pb2.Resource.Path = update.del_resource
                        await self.on_resource_changed(
                            resource_path.exporter_name, resource_path.group_name, resource_path.resource_name, {}
                        )
                    elif update_kind == "place":
                        place = update.place
                        await self.on_place_changed(place)
                    elif update_kind == "del_place":
                        place_name = update.del_place
                        await self.on_place_deleted(place_name)
                    else:
                        logging.warning("unknown update from coordinator! %s", update_kind)
                if out_msg.HasField("sync"):
                    event = self.sync_events.pop(out_msg.sync.id)
                    event.set()
        except grpc.aio.AioRpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                if got_message:
                    logging.error("coordinator became unavailable: %s", e.details())
                else:
                    logging.error("coordinator is unavailable: %s", e.details())
            else:
                logging.exception("unexpected grpc error in coordinator message pump task")
        except Exception:
            logging.exception("error in coordinator message pump task")
        finally:
            self.stopping.set()
            self.out_queue.put_nowait(None)  # let the sender side exit gracefully
            self.cancel_pending_syncs()

    async def on_resource_changed(self, exporter, group_name, resource_name, resource):
        group = self.resources.setdefault(exporter, {}).setdefault(group_name, {})
        # Do not replace the ResourceEntry object, as other components may keep
        # a reference to it and want to see changes.
        if resource_name not in group:
            old = None
            group[resource_name] = ResourceEntry(resource)
        else:
            old = group[resource_name].data
            group[resource_name].data = resource
        if self.monitor:
            if "cls" in resource and not old:
                print(f"Resource {exporter}/{group_name}/{resource['cls']}/{resource_name} created: {resource}")
            elif "cls" in resource and old:
                print(f"Resource {exporter}/{group_name}/{resource['cls']}/{resource_name} changed:")
                for k, v_old, v_new in diff_dict(flat_dict(old), flat_dict(resource)):
                    print(f"  {k}: {v_old} -> {v_new}")
            else:
                print(f"Resource {exporter}/{group_name}/???/{resource_name} deleted")

    async def on_place_changed(self, place_pb2: labgrid_coordinator_pb2.Place):
        name = place_pb2.name

        if name not in self.places:
            self.places[name] = Place.from_pb2(place_pb2)
            if self.monitor:
                print(f"Place {name} created: {place_pb2}")
        else:
            place = self.places[name]
            old = flat_dict(place.asdict())
            place.update_from_pb2(place_pb2)
            new = flat_dict(place.asdict())
            if self.monitor:
                print(f"Place {name} changed:")
                for k, v_old, v_new in diff_dict(old, new):
                    print(f"  {k}: {v_old} -> {v_new}")

    async def on_place_deleted(self, name: str):
        del self.places[name]
        if self.monitor:
            print(f"Place {name} deleted")

    async def do_monitor(self):
        self.monitor = True
        await self.stopping.wait()

    async def complete(self):
        if self.args.type == "resources":
            for exporter, groups in sorted(self.resources.items()):
                for group_name, group in sorted(groups.items()):
                    for _, resource in sorted(group.items()):
                        print(f"{exporter}/{group_name}/{resource.cls}")
        elif self.args.type == "places":
            for name in sorted(self.places.keys()):
                print(name)
        elif self.args.type == "matches":
            place = self.get_place()
            for match in place.matches:
                print(repr(match))
        elif self.args.type == "match-names":
            place = self.get_place()
            match_names = {match.rename for match in place.matches if match.rename is not None}
            print("\n".join(match_names))

    def _get_places_by_resource(self, resource_path):
        """Yield Place objects that match the given resource path"""
        for place in self.places.values():
            match = place.getmatch(resource_path)
            if match:
                yield place

    async def print_resources(self):
        """Print out the resources"""
        try:
            match = ResourceMatch.fromstr(self.args.match) if self.args.match else None
        except ValueError as e:
            raise UserError(str(e)) from e

        # filter self.resources according to the arguments
        nested = lambda: defaultdict(nested)
        filtered = nested()
        for exporter, groups in sorted(self.resources.items()):
            if self.args.exporter and exporter != self.args.exporter:
                continue
            for group_name, group in sorted(groups.items()):
                for resource_name, resource in sorted(group.items()):
                    if not (resource.avail or self.args.verbose >= 2):
                        continue
                    if self.args.acquired and resource.acquired is None:
                        continue
                    if match and not match.ismatch((exporter, group_name, resource.cls, resource_name)):
                        continue

                    filtered[exporter][group_name][resource_name] = resource

        # print the filtered resources
        if self.args.verbose and not self.args.sort_by_matched_place_change:
            for exporter, groups in sorted(filtered.items()):
                print(f"Exporter '{exporter}':")
                for group_name, group in sorted(groups.items()):
                    print(f"  Group '{group_name}' ({exporter}/{group_name}/*):")
                    for resource_name, resource in sorted(group.items()):
                        print(
                            "    Resource '{res}' ({exporter}/{group}/{res_cls}[/{res}]):".format(
                                res=resource_name, exporter=exporter, group=group_name, res_cls=resource.cls
                            )
                        )
                        print(indent(pformat(resource.asdict()), prefix="      "))
        else:
            results = []
            for exporter, groups in sorted(filtered.items()):
                for group_name, group in sorted(groups.items()):
                    for resource_name, resource in sorted(group.items()):
                        if self.args.sort_by_matched_place_change:
                            resource_path = (exporter, group_name, resource.cls, resource_name)
                            places = list(self._get_places_by_resource(resource_path))
                            # most recently changed place last
                            places = sorted(places, key=lambda p: p.changed, reverse=True)
                        else:
                            places = None

                        results.append((places, exporter, group_name, resource.cls))

            results = sorted(results, key=lambda res: res[0][0].changed if res[0] else 0)

            for places, exporter, group_name, resource_cls in results:
                if self.args.sort_by_matched_place_change:
                    places_strs = [f"{p.name}: {datetime.fromtimestamp(p.changed):%Y-%m-%d}" for p in places]
                    places_info = ", ".join(places_strs) if places_strs else "not used by any place"

                else:
                    places_info = None

                line = f"{exporter}/{group_name}/{resource_cls}"
                if places_info is not None:
                    print(f"{line:<50s} {places_info}")
                else:
                    print(line)

    async def print_places(self):
        """Print out the places"""
        if self.args.sort_last_changed:
            places = sorted(self.places.items(), key=lambda p: p[1].changed)
        else:
            places = sorted(self.places.items())

        for name, place in places:
            if self.args.acquired and place.acquired is None:
                continue
            if self.args.released and place.acquired:
                continue
            if self.args.verbose:
                print(f"Place '{name}':")
                place.show(level=1)
            else:
                line = f"{name}"

                if place.aliases:
                    line += f" ({' '.join(place.aliases)})"

                print(f"{line:<40s} {place.comment}")

    def print_who(self):
        """Print acquired places by user"""
        result = ["User Host Place Changed".split()]
        if self.args.show_exporters:
            result[0].append("Exporters")

        for name, place in self.places.items():
            if place.acquired is None:
                continue
            host, user = place.acquired.split("/")
            result.append([user, host, name, str(datetime.fromtimestamp(place.changed))])
            if self.args.show_exporters:
                exporters = {resource_path[0] for resource_path in place.acquired_resources}
                result[-1].append(", ".join(sorted(exporters)))
        result.sort()

        widths = [max(map(len, c)) for c in zip(*result)]
        layout = []
        for i, w in enumerate(widths):
            layout.append("{%i:<%is}" % (i, w))
        layout = "  ".join(layout)

        for entry in result:
            print(layout.format(*entry))

    def _match_places(self, pattern):
        """search for substring matches of pattern in place names and aliases

        Aliases in the format "namespace:alias" only match if the namespace is
        the current user name.
        """
        result = set()

        # reservation token lookup
        token = None
        if pattern.startswith("+"):
            token = pattern[1:]
            if not token:
                token = os.environ.get("LG_TOKEN", None)
            if not token:
                return []
            for name, place in self.places.items():
                if place.reservation == token:
                    result.add(name)
            if not result:
                raise UserError(f"reservation token {token} matches nothing")
            return list(result)

        # name and alias lookup
        for name, place in self.places.items():
            if pattern in name:
                result.add(name)
            for alias in place.aliases:
                if ":" in alias:
                    namespace, alias = alias.split(":", 1)
                    if namespace != self.getuser():
                        continue
                    if alias == pattern:  # prefer user namespace
                        return [name]
                if pattern in alias:
                    result.add(name)
        return list(result)

    def _check_allowed(self, place):
        if not place.acquired:
            raise UserError(f"place {place.name} is not acquired")
        if f"{self.gethostname()}/{self.getuser()}" not in place.allowed:
            host, user = place.acquired.split("/")
            if user != self.getuser():
                raise UserError(
                    f"place {place.name} is not acquired by your user, acquired by {user}. To work simultaneously, {user} can execute labgrid-client -p {place.name} allow {self.gethostname()}/{self.getuser()}"
                )
            if host != self.gethostname():
                raise UserError(
                    f"place {place.name} is not acquired on this computer, acquired on {host}. To allow this host, use labgrid-client -p {place.name} allow {self.gethostname()}/{self.getuser()} on the other host"
                )

    def get_place(self, place=None):
        pattern = place or self.args.place
        if pattern is None:
            raise UserError("place pattern not specified")
        places = self._match_places(pattern)
        if not places:
            raise UserError(f"place pattern {pattern} matches nothing")
        if pattern in places:
            return self.places[pattern]
        if len(places) > 1:
            raise UserError(f"pattern {pattern} matches multiple places ({', '.join(places)})")
        return self.places[places[0]]

    def get_place_names_from_env(self):
        """Returns a list of RemotePlace names found in the environment config."""
        places = []
        for role_config in self.env.config.get_targets().values():
            resources, _ = target_factory.normalize_config(role_config)
            remote_places = resources.get("RemotePlace", [])
            for place in remote_places:
                places.append(place)

        return places

    def get_idle_place(self, place=None):
        place = self.get_place(place)
        if place.acquired:
            err = "This operation requires the place to be idle."
            if place.acquired == self.gethostname() + "/" + self.getuser():
                raise UserError(f"{err} Please release {place.name}.")
            else:
                raise UserError(f"{err} Place {place.name} is acquired by {place.acquired}.")
        return place

    def get_acquired_place(self, place=None):
        place = self.get_place(place)
        self._check_allowed(place)
        return place

    async def print_place(self):
        """Print out the current place and related resources"""
        place = self.get_place()
        print(f"Place '{place.name}':")
        place.show(level=1)
        if place.acquired:
            for resource_path in place.acquired_resources:
                (exporter, group_name, cls, resource_name) = resource_path
                match = place.getmatch(resource_path)
                name = resource_name
                if match.rename:
                    name = match.rename
                try:
                    resource = self.resources[exporter][group_name][resource_name]
                except KeyError:
                    print(f"Orphaned resource '{name}' ({exporter}/{group_name}/{cls}/{resource_name})")
                    continue
                print(f"Acquired resource '{name}' ({exporter}/{group_name}/{resource.cls}/{resource_name}):")  # pylint: disable=line-too-long
                print(indent(pformat(resource.asdict()), prefix="  "))
                assert resource.cls == cls
        else:
            for exporter, groups in sorted(self.resources.items()):
                for group_name, group in sorted(groups.items()):
                    for resource_name, resource in sorted(group.items()):
                        resource_path = (exporter, group_name, resource.cls, resource_name)
                        match = place.getmatch(resource_path)
                        if match is None:
                            continue
                        name = resource_name
                        if match.rename:
                            name = match.rename
                        print(f"Matching resource '{name}' ({exporter}/{group_name}/{resource.cls}/{resource_name}):")  # pylint: disable=line-too-long
                        print(indent(pformat(resource.asdict()), prefix="  "))

    async def add_place(self):
        """Add a place to the coordinator"""
        name = self.args.place
        if not name:
            raise UserError("missing place name. Set with -p <place> or via env var LG_PLACE")

        request = labgrid_coordinator_pb2.AddPlaceRequest(name=name)
        try:
            await self.stub.AddPlace(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def del_place(self):
        """Delete a place from the coordinator"""
        place = self.get_idle_place()
        request = labgrid_coordinator_pb2.DeletePlaceRequest(name=place.name)
        try:
            await self.stub.DeletePlace(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def add_alias(self):
        """Add an alias for a place on the coordinator"""
        place = self.get_idle_place()
        alias = self.args.alias

        request = labgrid_coordinator_pb2.AddPlaceAliasRequest(placename=place.name, alias=alias)

        try:
            await self.stub.AddPlaceAlias(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def del_alias(self):
        """Delete an alias for a place from the coordinator"""
        place = self.get_idle_place()
        alias = self.args.alias

        request = labgrid_coordinator_pb2.DeletePlaceAliasRequest(placename=place.name, alias=alias)

        try:
            await self.stub.DeletePlaceAlias(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def set_comment(self):
        """Set the comment on a place"""
        place = self.get_place()
        comment = " ".join(self.args.comment)

        request = labgrid_coordinator_pb2.SetPlaceCommentRequest(placename=place.name, comment=comment)

        try:
            await self.stub.SetPlaceComment(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def set_tags(self):
        """Set the tags on a place"""
        place = self.get_place()
        tags = {}
        for pair in self.args.tags:
            try:
                k, v = pair.split("=")
            except ValueError:
                raise UserError(f"tag '{pair}' needs to match '<key>=<value>'")
            if not TAG_KEY.match(k):
                raise UserError(f"tag key '{k}' needs to match the rexex '{TAG_KEY.pattern}'")
            if not TAG_VAL.match(v):
                raise UserError(f"tag value '{v}' needs to match the rexex '{TAG_VAL.pattern}'")
            tags[k] = v

        request = labgrid_coordinator_pb2.SetPlaceTagsRequest(placename=place.name, tags=tags)

        try:
            await self.stub.SetPlaceTags(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def add_match(self):
        """Add a match for a place, making fuzzy matching available to the
        client"""
        place = self.get_idle_place()
        for pattern in self.args.patterns:
            if not 2 <= pattern.count("/") <= 3:
                raise UserError(f"invalid pattern format '{pattern}' (use 'exporter/group/cls/name')")
            if place.hasmatch(pattern.split("/")):
                print(f"pattern '{pattern}' exists, skipping", file=sys.stderr)
                continue

            request = labgrid_coordinator_pb2.AddPlaceMatchRequest(placename=place.name, pattern=pattern)

            try:
                await self.stub.AddPlaceMatch(request)
                await self.sync_with_coordinator()
            except grpc.aio.AioRpcError as e:
                raise ServerError(e.details())

    async def del_match(self):
        """Delete a match for a place"""
        place = self.get_idle_place()
        for pattern in self.args.patterns:
            if not 2 <= pattern.count("/") <= 3:
                raise UserError(f"invalid pattern format '{pattern}' (use 'exporter/group/cls/name')")
            if not place.hasmatch(pattern.split("/")):
                print(f"pattern '{pattern}' not found, skipping", file=sys.stderr)

            request = labgrid_coordinator_pb2.DeletePlaceMatchRequest(placename=place.name, pattern=pattern)

            try:
                await self.stub.DeletePlaceMatch(request)
                await self.sync_with_coordinator()
            except grpc.aio.AioRpcError as e:
                raise ServerError(e.details())

    async def add_named_match(self):
        """Add a named match for a place.

        Fuzzy matching is not allowed to avoid accidental names conflicts."""
        place = self.get_idle_place()
        pattern = self.args.pattern
        name = self.args.name
        if not 2 <= pattern.count("/") <= 3:
            raise UserError(f"invalid pattern format '{pattern}' (use 'exporter/group/cls/name')")
        if "*" in pattern:
            raise UserError(f"invalid pattern '{pattern}' ('*' not allowed for named matches)")

        request = labgrid_coordinator_pb2.AddPlaceMatchRequest(placename=place.name, pattern=pattern, rename=name)

        try:
            await self.stub.AddPlaceMatch(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    def check_matches(self, place):
        resources = []
        for exporter, groups in self.resources.items():
            for group_name, group in groups.items():
                for resource_name, resource in group.items():
                    resource_path = (exporter, group_name, resource.cls, resource_name)
                    resources.append(resource_path)

        match = place.unmatched(resources)
        if match:
            raise UserError(f"Match {match} has no matching remote resource")

    async def acquire(self):
        errors = []
        places = self.get_place_names_from_env() if self.env else [self.args.place]
        for place in places:
            try:
                await self._acquire_place(place)
            except Error as e:
                errors.append(e)

        if errors:
            if len(errors) == 1:
                raise errors[0]
            raise ErrorGroup("Multiple errors occurred during acquire", errors)

    async def _acquire_place(self, place):
        """Acquire a place, marking it unavailable for other clients"""
        place = self.get_place(place)
        if place.acquired:
            host, user = place.acquired.split("/")
            allowhelp = f"'labgrid-client -p {place.name} allow {self.gethostname()}/{self.getuser()}' on {host}."
            if self.getuser() == user:
                if self.gethostname() == host:
                    raise UserError(f"You have already acquired place {place.name}.")
                else:
                    raise UserError(
                        f"You have already acquired place {place.name} on {host}. To work simultaneously, execute {allowhelp}"
                    )
            else:
                raise UserError(
                    f"Place {place.name} is already acquired by {place.acquired}. To work simultaneously, {user} can execute {allowhelp}"
                )
        if not self.args.allow_unmatched:
            self.check_matches(place)

        request = labgrid_coordinator_pb2.AcquirePlaceRequest(placename=place.name)

        try:
            await self.stub.AcquirePlace(request)
            await self.sync_with_coordinator()
            print(f"acquired place {place.name}")
        except grpc.aio.AioRpcError as e:
            # check potential failure causes
            for exporter, groups in sorted(self.resources.items()):
                for group_name, group in sorted(groups.items()):
                    for resource_name, resource in sorted(group.items()):
                        resource_path = (exporter, group_name, resource.cls, resource_name)
                        if not resource.acquired:
                            continue
                        match = place.getmatch(resource_path)
                        if match is None:
                            continue
                        name = resource_name
                        if match.rename:
                            name = match.rename
                        print(
                            f"Matching resource '{name}' ({exporter}/{group_name}/{resource.cls}/{resource_name}) already acquired by place '{resource.acquired}'"
                        )  # pylint: disable=line-too-long

            raise ServerError(e.details())

    async def release(self):
        errors = []
        places = self.get_place_names_from_env() if self.env else [self.args.place]
        for place in places:
            try:
                await self._release_place(place)
            except Error as e:
                errors.append(e)

        if errors:
            if len(errors) == 1:
                raise errors[0]
            raise ErrorGroup("Multiple errors occurred during release", errors)

    async def _release_place(self, place):
        """Release a previously acquired place"""
        place = self.get_place(place)
        if not place.acquired:
            raise UserError(f"place {place.name} is not acquired")
        _, user = place.acquired.split("/")
        if user != self.getuser():
            if not self.args.kick:
                raise UserError(
                    f"place {place.name} is acquired by a different user ({place.acquired}), use --kick if you are sure"
                )  # pylint: disable=line-too-long
            print(f"warning: kicking user ({place.acquired})")

        request = labgrid_coordinator_pb2.ReleasePlaceRequest(placename=place.name)

        try:
            await self.stub.ReleasePlace(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

        print(f"released place {place.name}")

    async def release_from(self):
        """Release a place, but only if acquired by a specific user"""
        place = self.get_place()
        if not place.acquired:
            raise UserError(f"place {place.name} is not acquired")

        request = labgrid_coordinator_pb2.ReleasePlaceRequest(placename=place.name, fromuser=self.args.acquired)

        try:
            await self.stub.ReleasePlace(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

        print(f"{self.args.acquired} has released place {place.name}")

    async def allow(self):
        """Allow another use access to a previously acquired place"""
        place = self.get_acquired_place()
        if "/" not in self.args.user:
            raise UserError(f"user {self.args.user} must be in <host>/<username> format")
        request = labgrid_coordinator_pb2.AllowPlaceRequest(placename=place.name, user=self.args.user)

        try:
            await self.stub.AllowPlace(request)
            await self.sync_with_coordinator()
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

        print(f"allowed {self.args.user} for place {place.name}")

    def get_target_resources(self, place):
        self._check_allowed(place)
        resources = {}
        for resource_path in place.acquired_resources:
            match = place.getmatch(resource_path)
            (exporter, group_name, cls, resource_name) = resource_path
            name = resource_name
            if match.rename:
                name = match.rename
            try:
                resources[(name, cls)] = self.resources[exporter][group_name][resource_name]
            except KeyError:
                raise ServerError(f"place {place} has an orphaned resource (exporter {exporter} disconnected?)")

        return resources

    def get_target_config(self, place):
        config = {}
        resources = config["resources"] = []
        for (name, _), resource in self.get_target_resources(place).items():
            args = OrderedDict()
            if name != resource.cls:
                args["name"] = name
            args.update(resource.args)
            resources.append({resource.cls: args})
        return config

    def print_env(self):
        place = self.get_acquired_place()
        env = {"targets": {place.name: self.get_target_config(place)}}
        print(dump(env))

    def _prepare_manager(self):
        manager = RemotePlaceManager.get()
        manager.session = self
        manager.loop = self.loop

    def _get_target(self, place):
        self._prepare_manager()
        target = None
        if self.env:
            if self.role is None:
                self.role = find_role_by_place(self.env.config.get_targets(), place.name)
                if self.role is not None:
                    print(f"Selected role {self.role} from configuration file")
            target = self.env.get_target(self.role)
        if target:
            if self.args.state:
                strategy = target.get_driver("Strategy")
                if self.args.initial_state:
                    print(f"Setting initial state to {self.args.initial_state}")
                    strategy.force(self.args.initial_state)
                print(f"Transitioning into state {self.args.state}")
                strategy.transition(self.args.state)
                # deactivate console drivers so we are able to connect with microcom later
                try:
                    con = target.get_active_driver("ConsoleProtocol")
                    target.deactivate(con)
                except NoDriverFoundError:
                    pass
        else:
            target = Target(place.name, env=self.env)
            RemotePlace(target, name=place.name)
        return target

    def _get_driver_or_new(self, target, cls, *, name=None, activate=True):
        """
        Helper function trying to get an active driver. If no such driver
        exists, instanciates a new driver.
        Driver instanciation works only for drivers without special kwargs.

        Arguments:
        target -- target to operate on
        cls -- driver-class to retrieve active or instanciate new driver from
        name -- optional name to use as a filter
        activate -- activate the driver (default True)
        """
        try:
            return target.get_driver(cls, name=name, activate=activate)
        except NoDriverFoundError:
            if isinstance(cls, str):
                cls = target_factory.class_from_string(cls)

            if name is not None:
                # set name in binding map for unique bindings
                try:
                    [unique_binding_key] = cls.bindings
                    target.set_binding_map({unique_binding_key: name})
                except ValueError:
                    raise NotImplementedError("Multiple bindings not implemented for named resources")

            drv = cls(target, name=name)
            if activate:
                target.activate(drv)
            return drv

    def power(self):
        place = self.get_acquired_place()
        action = self.args.action
        delay = self.args.delay
        name = self.args.name
        target = self._get_target(place)
        from ..resource.power import NetworkPowerPort, PDUDaemonPort
        from ..resource.remote import NetworkUSBPowerPort, NetworkSiSPMPowerPort, NetworkSysfsGPIO
        from ..resource import TasmotaPowerPort, NetworkYKUSHPowerPort

        drv = None
        try:
            drv = target.get_driver("PowerProtocol", name=name)
        except NoDriverFoundError:
            for resource in target.resources:
                if name and resource.name != name:
                    continue
                if isinstance(resource, NetworkPowerPort):
                    drv = self._get_driver_or_new(target, "NetworkPowerDriver", name=name)
                elif isinstance(resource, NetworkUSBPowerPort):
                    drv = self._get_driver_or_new(target, "USBPowerDriver", name=name)
                elif isinstance(resource, NetworkSiSPMPowerPort):
                    drv = self._get_driver_or_new(target, "SiSPMPowerDriver", name=name)
                elif isinstance(resource, PDUDaemonPort):
                    drv = self._get_driver_or_new(target, "PDUDaemonDriver", name=name)
                elif isinstance(resource, TasmotaPowerPort):
                    drv = self._get_driver_or_new(target, "TasmotaPowerDriver", name=name)
                elif isinstance(resource, NetworkYKUSHPowerPort):
                    drv = self._get_driver_or_new(target, "YKUSHPowerDriver", name=name)
                elif isinstance(resource, NetworkSysfsGPIO):
                    self._get_driver_or_new(target, "GpioDigitalOutputDriver", name=name)
                    drv = self._get_driver_or_new(target, "DigitalOutputPowerDriver", name=name)
                if drv:
                    break

        if not drv:
            raise UserError("target has no compatible resource available")
        if delay is not None:
            drv.delay = delay
        res = getattr(drv, action)()
        if action == "get":
            print(f"power{' ' + name if name else ''} for place {place.name} is {'on' if res else 'off'}")

    def digital_io(self):
        place = self.get_acquired_place()
        action = self.args.action
        name = self.args.name
        target = self._get_target(place)
        from ..resource import (
            ModbusTCPCoil,
            OneWirePIO,
            HttpDigitalOutput,
            WaveshareModbusTCPCoil,
            Eth008DigitalOutput,
        )
        from ..resource.remote import NetworkDeditecRelais8, NetworkSysfsGPIO, NetworkLXAIOBusPIO, NetworkHIDRelay

        drv = None
        try:
            drv = target.get_driver("DigitalOutputProtocol", name=name)
        except NoDriverFoundError:
            for resource in target.resources:
                if name and resource.name != name:
                    continue
                if isinstance(resource, WaveshareModbusTCPCoil):
                    drv = self._get_driver_or_new(target, "WaveShareModbusCoilDriver", name=name)
                elif isinstance(resource, ModbusTCPCoil):
                    drv = self._get_driver_or_new(target, "ModbusCoilDriver", name=name)
                elif isinstance(resource, Eth008DigitalOutput):
                    drv = self._get_driver_or_new(target, "Eth008DigitalOutputDriver", name=name)
                elif isinstance(resource, OneWirePIO):
                    drv = self._get_driver_or_new(target, "OneWirePIODriver", name=name)
                elif isinstance(resource, HttpDigitalOutput):
                    drv = self._get_driver_or_new(target, "HttpDigitalOutputDriver", name=name)
                elif isinstance(resource, NetworkDeditecRelais8):
                    drv = self._get_driver_or_new(target, "DeditecRelaisDriver", name=name)
                elif isinstance(resource, NetworkSysfsGPIO):
                    drv = self._get_driver_or_new(target, "GpioDigitalOutputDriver", name=name)
                elif isinstance(resource, NetworkLXAIOBusPIO):
                    drv = self._get_driver_or_new(target, "LXAIOBusPIODriver", name=name)
                elif isinstance(resource, NetworkHIDRelay):
                    drv = self._get_driver_or_new(target, "HIDRelayDriver", name=name)
                if drv:
                    break

        if not drv:
            raise UserError("target has no compatible resource available")
        if action == "get":
            print(f"digital IO{' ' + name if name else ''} for place {place.name} is {'high' if drv.get() else 'low'}")
        elif action == "high":
            drv.set(True)
        elif action == "low":
            drv.set(False)

    async def _console(self, place, target, timeout, *, logfile=None, loop=False, listen_only=False):
        name = self.args.name
        from ..resource import NetworkSerialPort

        resource = target.get_resource(NetworkSerialPort, name=name, wait_avail=False)

        # async await resources
        timeout = Timeout(timeout)
        while True:
            target.update_resources()
            if resource.avail or (not loop and timeout.expired):
                break
            await asyncio.sleep(0.1)

        # use zero timeout to prevent blocking sleeps
        target.await_resources([resource], timeout=0.0)

        if not place.acquired:
            print("place released")
            return 255

        host, port = proxymanager.get_host_and_port(resource)

        # check for valid resources
        assert port is not None, "Port is not set"

        microcom_bin = shutil.which("microcom")

        if microcom_bin is not None:
            call = [microcom_bin, "-s", str(resource.speed), "-t", f"{host}:{port}"]

            if listen_only:
                call.append("--listenonly")

            if logfile:
                call.append(f"--logfile={logfile}")
        else:
            call = ["telnet", host, str(port)]

            logging.info("microcom not available, using telnet instead")

            if listen_only:
                logging.warning("--listenonly option not supported by telnet, ignoring")

            if logfile:
                logging.warning("--logfile option not supported by telnet, ignoring")

        print(f"connecting to {resource} calling {' '.join(call)}")
        try:
            p = await asyncio.create_subprocess_exec(*call)
        except FileNotFoundError as e:
            raise ServerError(f"failed to execute remote console command: {e}")
        while p.returncode is None:
            try:
                await asyncio.wait_for(p.wait(), 1.0)
            except asyncio.TimeoutError:
                # subprocess is still running
                pass

            try:
                self._check_allowed(place)
            except UserError:
                p.terminate()
                try:
                    await asyncio.wait_for(p.wait(), 1.0)
                except asyncio.TimeoutError:
                    # try harder
                    p.kill()
                    await asyncio.wait_for(p.wait(), 1.0)
                raise
        if p.returncode:
            print("connection lost", file=sys.stderr)
        return p.returncode

    async def console(self, place, target):
        while True:
            res = await self._console(
                place, target, 10.0, logfile=self.args.logfile, loop=self.args.loop, listen_only=self.args.listenonly
            )
            # place released
            if res == 255:
                break
            if not self.args.loop:
                if res:
                    raise InteractiveCommandError("microcom error", res)
                break
            await asyncio.sleep(1.0)

    console.needs_target = True

    def dfu(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name
        if self.args.action == "download" and not self.args.filename:
            raise UserError("not enough arguments for dfu download")
        drv = self._get_driver_or_new(target, "DFUDriver", activate=False, name=name)
        drv.dfu.timeout = self.args.wait
        target.activate(drv)

        if self.args.action == "download":
            drv.download(self.args.altsetting, os.path.abspath(self.args.filename))
        if self.args.action == "detach":
            drv.detach(self.args.altsetting)
        if self.args.action == "list":
            drv.list()

    def fastboot(self):
        place = self.get_acquired_place()
        args = self.args.fastboot_args
        target = self._get_target(place)
        name = self.args.name

        drv = self._get_driver_or_new(target, "AndroidFastbootDriver", activate=False, name=name)
        drv.fastboot.timeout = self.args.wait
        target.activate(drv)

        try:
            action = args[0]
            if action == "flash":
                drv.flash(args[1], os.path.abspath(args[2]))
            elif action == "boot":
                args[1:] = map(os.path.abspath, args[1:])
                drv.boot(args[1])
            elif action == "oem" and args[1] == "exec":
                drv.run(" ".join(args[2:]))
            else:
                drv(*args)
        except IndexError:
            raise UserError("not enough arguments for fastboot action")
        except subprocess.CalledProcessError as e:
            raise UserError(str(e))

    def flashscript(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name

        drv = self._get_driver_or_new(target, "FlashScriptDriver", name=name)
        drv.flash(script=self.args.script, args=self.args.script_args)

    def bootstrap(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name
        from ..resource.remote import (
            NetworkMXSUSBLoader,
            NetworkIMXUSBLoader,
            NetworkRKUSBLoader,
            NetworkAlteraUSBBlaster,
        )
        from ..driver import OpenOCDDriver

        drv = None
        try:
            drv = target.get_driver("BootstrapProtocol", name=name)
        except NoDriverFoundError:
            for resource in target.resources:
                if name and resource.name != name:
                    continue
                if isinstance(resource, NetworkIMXUSBLoader):
                    drv = self._get_driver_or_new(target, "IMXUSBDriver", activate=False, name=name)
                    drv.loader.timeout = self.args.wait
                elif isinstance(resource, NetworkMXSUSBLoader):
                    drv = self._get_driver_or_new(target, "MXSUSBDriver", activate=False, name=name)
                    drv.loader.timeout = self.args.wait
                elif isinstance(resource, NetworkAlteraUSBBlaster):
                    args = dict(arg.split("=", 1) for arg in self.args.bootstrap_args)
                    try:
                        drv = target.get_driver("OpenOCDDriver", activate=False, name=name)
                    except NoDriverFoundError:
                        drv = OpenOCDDriver(target, name=name, **args)
                    drv.interface.timeout = self.args.wait
                elif isinstance(resource, NetworkRKUSBLoader):
                    drv = self._get_driver_or_new(target, "RKUSBDriver", activate=False, name=name)
                    drv.loader.timeout = self.args.wait
                if drv:
                    break

        if not drv:
            raise UserError("target has no compatible resource available")
        target.activate(drv)
        drv.load(self.args.filename)

    def sd_mux(self):
        place = self.get_acquired_place()
        action = self.args.action
        target = self._get_target(place)
        name = self.args.name
        from ..resource.remote import NetworkUSBSDMuxDevice, NetworkUSBSDWireDevice, NetworkUSBSDWire3Device

        drv = None
        for resource in target.resources:
            if name and resource.name != name:
                continue
            if isinstance(resource, NetworkUSBSDMuxDevice):
                drv = self._get_driver_or_new(target, "USBSDMuxDriver", name=name)
            elif isinstance(resource, NetworkUSBSDWireDevice):
                drv = self._get_driver_or_new(target, "USBSDWireDriver", name=name)
            elif isinstance(resource, NetworkUSBSDWire3Device):
                drv = self._get_driver_or_new(target, "USBSDWire3Driver", name=name)
            if drv:
                break

        if not drv:
            raise UserError("target has no compatible resource available")
        if action == "get":
            print(drv.get_mode())
        else:
            try:
                drv.set_mode(action)
            except ExecutionError as e:
                raise UserError(str(e))

    def usb_mux(self):
        place = self.get_acquired_place()
        name = self.args.name
        links = self.args.links
        if links == "off":
            links = []
        elif links == "host-dut+host-device":
            links = ["host-dut", "host-device"]
        else:
            links = [links]
        target = self._get_target(place)
        from ..resource.remote import NetworkLXAUSBMux

        drv = None
        for resource in target.resources:
            if name and resource.name != name:
                continue
            if isinstance(resource, NetworkLXAUSBMux):
                drv = self._get_driver_or_new(target, "LXAUSBMuxDriver", name=name)
                break

        if not drv:
            raise UserError("target has no compatible resource available")
        drv.set_links(links)

    def _get_ip(self, place):
        target = self._get_target(place)
        try:
            resource = target.get_resource("EthernetPort")
        except NoResourceFoundError:
            resource = target.get_resource("NetworkService")
            return resource.address

        matches = []
        for details in resource.extra.get("macs").values():
            ips = details.get("ips", [])
            if not ips:
                continue
            matches.append((details["timestamp"], ips))
        matches.sort()
        newest = matches[-1][1]
        if len(ips) > 1:
            print(f"multiple IPs found: {ips}", file=sys.stderr)
            return None
        return newest[0]

    def _get_ssh(self):
        place = self.get_acquired_place()
        target = self._get_target(place)

        try:
            drv = target.get_driver("SSHDriver", name=self.args.name)
            return drv
        except NoDriverFoundError:
            from ..resource import NetworkService

            try:
                resource = target.get_resource(NetworkService, name=self.args.name)
            except NoResourceFoundError:
                ip = self._get_ip(place)
                if not ip:
                    return
                resource = NetworkService(target, address=str(ip), username="root")

            drv = self._get_driver_or_new(target, "SSHDriver", name=resource.name)
            return drv

    def ssh(self):
        drv = self._get_ssh()

        res = drv.interact(self.args.leftover)
        if res:
            raise InteractiveCommandError("ssh error", res)

    def scp(self):
        drv = self._get_ssh()

        res = drv.scp(src=self.args.src, dst=self.args.dst)
        if res:
            raise InteractiveCommandError("scp error", res)

    def rsync(self):
        drv = self._get_ssh()

        res = drv.rsync(src=self.args.src, dst=self.args.dst, extra=self.args.leftover)
        if res:
            raise InteractiveCommandError("rsync error", res)

    def sshfs(self):
        drv = self._get_ssh()

        drv.sshfs(path=self.args.path, mountpoint=self.args.mountpoint)

    def forward(self):
        if not self.args.local and not self.args.remote:
            print("Nothing to forward", file=sys.stderr)
            return

        drv = self._get_ssh()

        with contextlib.ExitStack() as stack:
            for local, remote in self.args.local:
                localport = stack.enter_context(drv.forward_local_port(remote, localport=local))
                print(f"Forwarding local port {localport:d} to remote port {remote:d}")

            for local, remote in self.args.remote:
                stack.enter_context(drv.forward_remote_port(remote, local))
                print(f"Forwarding remote port {remote:d} to local port {local:d}")

            try:
                print("Waiting for CTRL+C...")
                while True:
                    signal.pause()
            except KeyboardInterrupt:
                print("Exiting...")

    def telnet(self):
        place = self.get_acquired_place()
        ip = self._get_ip(place)
        if not ip:
            return
        args = ["telnet", str(ip)]
        res = subprocess.call(args)
        if res:
            raise InteractiveCommandError("telnet error", res)

    def video(self):
        place = self.get_acquired_place()
        quality = self.args.quality
        controls = self.args.controls
        target = self._get_target(place)
        name = self.args.name
        from ..resource.httpvideostream import HTTPVideoStream
        from ..resource.udev import USBVideo
        from ..resource.remote import NetworkUSBVideo

        drv = None
        try:
            drv = target.get_driver("VideoProtocol", name=name)
        except NoDriverFoundError:
            for resource in target.resources:
                if name and resource.name != name:
                    continue
                if isinstance(resource, (USBVideo, NetworkUSBVideo)):
                    drv = self._get_driver_or_new(target, "USBVideoDriver", name=name)
                elif isinstance(resource, HTTPVideoStream):
                    drv = self._get_driver_or_new(target, "HTTPVideoDriver", name=name)
                if drv:
                    break
        if not drv:
            raise UserError("target has no compatible resource available")

        if quality == "list":
            default, variants = drv.get_qualities()
            for name, caps in variants:
                mark = "*" if default == name else " "
                print(f"{mark} {name:<10s} {caps:s}")
        else:
            res = drv.stream(quality, controls=controls)
            if res:
                raise InteractiveCommandError("gst-launch-1.0 error", res)

    def audio(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name
        drv = self._get_driver_or_new(target, "USBAudioInputDriver", name=name)
        res = drv.play()
        if res:
            raise InteractiveCommandError("gst-launch-1.0 error", res)

    def netns(self):
        place = self.get_acquired_place()
        with self._get_target(place) as target:
            name = self.args.name
            drv = self._get_driver_or_new(target, "RawNetworkInterfaceDriver", name=name)
            with drv.setup_netns(self.args.mac_address) as ns:
                for a in self.args.address:
                    ns.run(["ip", "addr", "add", str(a), "dev", ns.intf], check=True)

                if self.args.nsenter:
                    with open(self.args.nsenter, "w") as wrapper:
                        wrapper.write(
                            "\n".join(
                                [
                                    "#!/usr/bin/sh",
                                    " ".join(ns.prefix) + ' "$@"',
                                    "",
                                ]
                            )
                        )
                        os.fchmod(wrapper.fileno(), 0o755)
                    print(f"Use {self.args.nsenter} to enter the namespace")

                cmd = self.args.cmd
                if not cmd:
                    # Same behavior as nsenter with no command
                    cmd = os.environ.get("SHELL", "/bin/sh")

                p = ns.run(cmd)
                if p.returncode != 0:
                    raise InteractiveCommandError("netns command error", p.returncode)

    def _get_tmc(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name

        return self._get_driver_or_new(target, "USBTMCDriver", name=name)

    def tmc_command(self):
        drv = self._get_tmc()
        command = " ".join(self.args.command)
        if not command:
            raise UserError("no command given")
        if "?" in command:
            result = drv.query(command)
            print(result)
        else:
            drv.command(command)

    def tmc_query(self):
        drv = self._get_tmc()
        query = " ".join(self.args.query)
        if not query:
            raise UserError("no query given")
        result = drv.query(query)
        print(result)

    def tmc_screen(self):
        drv = self._get_tmc()
        action = self.args.action
        if action in ["show", "save"]:
            extension, data = drv.get_screenshot()
            filename = "tmc-screen_{0:%Y-%m-%d}_{0:%H:%M:%S}.{1}".format(datetime.now(), extension)
            with open(filename, "wb") as f:
                f.write(data)
            print(f"Saved as {filename}")
            if action == "show":
                subprocess.call(["xdg-open", filename])

    def tmc_channel(self):
        drv = self._get_tmc()
        channel = self.args.channel
        action = self.args.action
        if action == "info":
            data = drv.get_channel_info(channel)
        elif action == "values":
            data = drv.get_channel_values(channel)
        else:
            raise ValueError(f"unknown action {action}")

        for k, v in sorted(data.items()):
            print(f"{k:<16s} {str(v):<10s}")

    def write_files(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name
        drv = self._get_driver_or_new(target, "USBStorageDriver", activate=False, name=name)
        drv.storage.timeout = self.args.wait
        target.activate(drv)

        try:
            if self.args.partition == 0:
                self.args.partition = None

            if self.args.rename:
                if len(self.args.SOURCE) != 2:
                    self.args.parser.error("the following arguments are required: SOURCE DEST")

                drv.write_files(
                    [self.args.SOURCE[0]], self.args.SOURCE[1], self.args.partition, target_is_directory=False
                )
            else:
                drv.write_files(
                    self.args.SOURCE, self.args.target_directory, self.args.partition, target_is_directory=True
                )
        except subprocess.CalledProcessError as e:
            raise UserError(f"could not copy files to network usb storage: {e}")
        except FileNotFoundError as e:
            raise UserError(e)

    def write_image(self):
        place = self.get_acquired_place()
        target = self._get_target(place)
        name = self.args.name
        drv = self._get_driver_or_new(target, "USBStorageDriver", activate=False, name=name)
        drv.storage.timeout = self.args.wait
        target.activate(drv)

        try:
            drv.write_image(
                self.args.filename,
                partition=self.args.partition,
                skip=self.args.skip,
                seek=self.args.seek,
                mode=self.args.write_mode,
            )
        except subprocess.CalledProcessError as e:
            raise UserError(f"could not write image to network usb storage: {e}")
        except FileNotFoundError as e:
            raise UserError(e)

    async def create_reservation(self):
        prio = self.args.prio

        fltr = {}
        for pair in self.args.filters:
            try:
                k, v = pair.split("=")
            except ValueError:
                raise UserError(f"'{pair}' is not a valid filter (must contain a '=')")
            if not TAG_KEY.match(k):
                raise UserError(f"Key '{k}' in filter '{pair}' is invalid")
            if not TAG_VAL.match(v):
                raise UserError(f"Value '{v}' in filter '{pair}' is invalid")
            fltr[k] = v

        fltrs = {
            "main": labgrid_coordinator_pb2.Reservation.Filter(filter=fltr),
        }

        request = labgrid_coordinator_pb2.CreateReservationRequest(filters=fltrs, prio=prio)

        try:
            response: labgrid_coordinator_pb2.CreateReservationResponse = await self.stub.CreateReservation(request)
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

        res = Reservation.from_pb2(response.reservation)
        if self.args.shell:
            print(f"export LG_TOKEN={res.token}")
        else:
            print(f"Reservation '{res.token}':")
            res.show(level=1)
        if self.args.wait:
            if not self.args.shell:
                print("Waiting for allocation...")
            await self._wait_reservation(res.token, verbose=False)

    async def cancel_reservation(self):
        token: str = self.args.token

        request = labgrid_coordinator_pb2.CancelReservationRequest(token=token)

        try:
            await self.stub.CancelReservation(request)
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

    async def _wait_reservation(self, token: str, verbose=True):
        while True:
            request = labgrid_coordinator_pb2.PollReservationRequest(token=token)

            try:
                response: labgrid_coordinator_pb2.PollReservationResponse = await self.stub.PollReservation(request)
            except grpc.aio.AioRpcError as e:
                raise ServerError(e.details())

            res = Reservation.from_pb2(response.reservation)
            if verbose:
                res.show()
            if res.state is ReservationState.waiting:
                await asyncio.sleep(1.0)
            else:
                break

    async def wait_reservation(self):
        token = self.args.token
        await self._wait_reservation(token)

    async def print_reservations(self):
        request = labgrid_coordinator_pb2.GetReservationsRequest()

        try:
            response: labgrid_coordinator_pb2.GetReservationsResponse = await self.stub.GetReservations(request)
            reservations = [Reservation.from_pb2(x) for x in response.reservations]
        except grpc.aio.AioRpcError as e:
            raise ServerError(e.details())

        for res in sorted(reservations, key=lambda x: (-x.prio, x.created)):
            print(f"Reservation '{res.token}':")
            res.show(level=1)

    async def export(self, place, target):
        exported = target.export()
        exported["LG__CLIENT_PID"] = str(os.getpid())
        if self.args.format is ExportFormat.SHELL:
            lines = []
            for k, v in sorted(exported.items()):
                lines.append(f"{k}={shlex.quote(v)}")
            data = "\n".join(lines)
        elif self.args.format is ExportFormat.SHELL_EXPORT:
            lines = []
            for k, v in sorted(exported.items()):
                lines.append(f"export {k}={shlex.quote(v)}")
            data = "\n".join(lines) + "\n"
        elif self.args.format is ExportFormat.JSON:
            data = json.dumps(exported)
        else:
            raise NotImplementedError(f"unsupported format {self.args.format}")

        if self.args.filename == "-":
            sys.stdout.write(data)
        else:
            atomic_replace(self.args.filename, data.encode())
            print(f"Exported to {self.args.filename}", file=sys.stderr)
        try:
            print("Waiting for CTRL+C or SIGTERM...", file=sys.stderr)
            while True:
                await asyncio.sleep(1.0)
        except GeneratorExit:
            print("Exiting...\n", file=sys.stderr)

    export.needs_target = True

    def print_version(self):
        print(labgrid_version())


_loop: ContextVar["asyncio.AbstractEventLoop | None"] = ContextVar("_loop", default=None)


def ensure_event_loop(external_loop=None):
    """Get the event loop for this thread, or create a new event loop."""
    # get stashed loop
    loop = _loop.get()

    # ignore closed stashed loop
    if loop and loop.is_closed():
        loop = None

    if external_loop:
        # if a loop is stashed, expect it to be the same as the external one
        if loop:
            assert loop is external_loop
        _loop.set(external_loop)
        return external_loop

    # return stashed loop
    if loop:
        return loop

    try:
        # if called from async code, try to get current's thread loop
        loop = asyncio.get_running_loop()
    except RuntimeError:
        # no previous, external or running loop found, create a new one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    # stash it
    _loop.set(loop)
    return loop


def start_session(
    address: str, *, extra: Dict[str, Any] = None, debug: bool = False, loop: "asyncio.AbstractEventLoop | None" = None
):
    """
    Starts a ClientSession.

    Args:
        address: coordinator address as HOST[:PORT], PORT defaults to 20408
        extra: additional kwargs for ClientSession
        debug: set debug mode of the event loop
        loop: explicit event loop to use (otherwise a previously stashed loop,
              if retrievable the current thread's loop or a new loop is used)
    """
    loop = ensure_event_loop(loop)

    if extra is None:
        extra = {}

    if debug:
        loop.set_debug(True)

    address = proxymanager.get_grpc_address(address, default_port=20408)

    session = ClientSession(address, loop, **extra)
    loop.run_until_complete(session.start())
    return session


def find_role_by_place(config, place):
    for role, role_config in config.items():
        resources, _ = target_factory.normalize_config(role_config)
        remote_places = resources.get("RemotePlace", {})
        remote_place = remote_places.get(place)
        if remote_place:
            return role
    return None


def find_any_role_with_place(config):
    for role, role_config in config.items():
        resources, _ = target_factory.normalize_config(role_config)
        remote_places = resources.get("RemotePlace", {})
        for place in remote_places:
            return (role, place)
    return None, None


class LocalPort(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super().__init__(option_strings, dest, nargs=None, default=[], **kwargs)

    def __call__(self, parser, namespace, value, option_string):
        if ":" in value:
            (local, remote) = value.split(":")
            local = int(local)
            remote = int(remote)
        else:
            local = None
            remote = int(value)

        v = getattr(namespace, self.dest, [])
        v.append((local, remote))
        setattr(namespace, self.dest, v)


class RemotePort(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super().__init__(option_strings, dest, nargs=None, default=[], **kwargs)

    def __call__(self, parser, namespace, value, option_string):
        (remote, local) = value.split(":")
        remote = int(remote)
        local = int(local)

        v = getattr(namespace, self.dest, [])
        v.append((local, remote))
        setattr(namespace, self.dest, v)


class ExportFormat(enum.Enum):
    SHELL = "shell"
    SHELL_EXPORT = "shell-export"
    JSON = "json"

    def __str__(self):
        return self.value


class AutoProgramArgumentParser(argparse.ArgumentParser):
    """Works around sphinxcontrib.autoprogram shortcomings."""

    class _ActionWrapper:
        """
        Wraps argparse's special private action object registered for "parsers", see:
        https://docs.python.org/3.14/library/argparse.html#argparse.ArgumentParser.add_subparsers
        https://docs.python.org/3.13/library/argparse.html#argparse.ArgumentParser.register
        """

        def __init__(self, action):
            self.action = action

        def add_parser(self, name, **kwargs):
            # aliases are not supported by autoprogram, they lead to duplicate entries, so
            # show them as "command subcommand|alias --option" instead
            aliases = kwargs.pop("aliases", [])
            if aliases:
                name = "|".join([name] + list(aliases))

            # "help" text is ignored by autoprogram, move to "description" instead
            if "description" not in kwargs and "help" in kwargs:
                kwargs["description"] = kwargs.pop("help")

            # add_parser() is the only part of the action object that's not an implementation detail
            return self.action.add_parser(name, **kwargs)

    def __init__(self, *args, **kwargs):
        # print help texts in fixed width
        os.environ["COLUMNS"] = "80"

        # hide --help, -h
        kwargs.setdefault("add_help", False)

        super().__init__(*args, **kwargs)

    def add_subparsers(self, **kwargs):
        action = super().add_subparsers(**kwargs)
        return self._ActionWrapper(action)

    def add_argument(self, *args, **kwargs):
        # do not show ranges
        if isinstance(kwargs.get("choices"), range):
            del kwargs["choices"]

        return super().add_argument(*args, **kwargs)


def get_parser(auto_doc_mode=False) -> "argparse.ArgumentParser | AutoProgramArgumentParser":
    # use custom parser for sphinxcontrib.autoprogram
    parser = AutoProgramArgumentParser() if auto_doc_mode else argparse.ArgumentParser()

    parser.add_argument(
        "-x",
        "--coordinator",
        metavar="ADDRESS",
        type=str,
        help="coordinator HOST[:PORT] (default: value from env variable LG_COORDINATOR, otherwise 127.0.0.1:20408)",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="env config file (default: value from env variable LG_ENV)",
    )
    parser.add_argument("-p", "--place", type=str, help="place name/alias (default: value from env variable LG_PLACE)")
    parser.add_argument(
        "-s",
        "--state",
        type=str,
        help="strategy state to switch into before command (default: value from env varibale LG_STATE)",
    )
    parser.add_argument(
        "-i",
        "--initial-state",
        type=str,
        help="strategy state to force into before switching to desired state",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", default=False, help="enable debug mode (show python tracebacks)"
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-P", "--proxy", type=str, help="proxy connections via given ssh host")
    subparsers = parser.add_subparsers(
        dest="command",
        title="available subcommands",
        metavar="COMMAND",
    )

    # if the argparse object is to be used for manpage generation, hide some subcommands
    if not auto_doc_mode:
        subparser = subparsers.add_parser("help")

        subparser = subparsers.add_parser("complete")
        subparser.add_argument("type", choices=["resources", "places", "matches", "match-names"])
        subparser.set_defaults(func=ClientSession.complete)

    subparser = subparsers.add_parser("monitor", help="monitor events from the coordinator")
    subparser.set_defaults(func=ClientSession.do_monitor)

    subparser = subparsers.add_parser("resources", aliases=("r",), help="list available resources")
    subparser.add_argument("-a", "--acquired", action="store_true")
    subparser.add_argument("-e", "--exporter")
    subparser.add_argument(
        "--sort-by-matched-place-change",
        action="store_true",
        help="sort by matched place's changed date (oldest first) and show place and date",
    )  # pylint: disable=line-too-long
    subparser.add_argument("match", nargs="?")
    subparser.set_defaults(func=ClientSession.print_resources)

    subparser = subparsers.add_parser("places", aliases=("p",), help="list available places")
    subparser.add_argument("-a", "--acquired", action="store_true")
    subparser.add_argument("-r", "--released", action="store_true")
    subparser.add_argument("--sort-last-changed", action="store_true", help="sort by last changed date (oldest first)")
    subparser.set_defaults(func=ClientSession.print_places)

    subparser = subparsers.add_parser("who", help="list acquired places by user")
    subparser.add_argument(
        "-e", "--show-exporters", action="store_true", help="show exporters currently used by each place"
    )
    subparser.set_defaults(func=ClientSession.print_who)

    subparser = subparsers.add_parser("show", help="show a place and related resources")
    subparser.set_defaults(func=ClientSession.print_place)

    subparser = subparsers.add_parser(
        "create",
        help="add a new place with the name specified via --place or the LG_PLACE environment variable",
    )
    subparser.set_defaults(func=ClientSession.add_place)

    subparser = subparsers.add_parser("delete", help="delete an existing place")
    subparser.set_defaults(func=ClientSession.del_place)

    subparser = subparsers.add_parser("add-alias", help="add an alias to a place")
    subparser.add_argument("alias")
    subparser.set_defaults(func=ClientSession.add_alias)

    subparser = subparsers.add_parser("del-alias", help="delete an alias from a place")
    subparser.add_argument("alias")
    subparser.set_defaults(func=ClientSession.del_alias)

    subparser = subparsers.add_parser("set-comment", help="update the place comment")
    subparser.add_argument("comment", nargs="+")
    subparser.set_defaults(func=ClientSession.set_comment)

    subparser = subparsers.add_parser("set-tags", help="update the place tags")
    subparser.add_argument("tags", metavar="KEY=VALUE", nargs="+", help="use an empty value for deletion")
    subparser.set_defaults(func=ClientSession.set_tags)

    subparser = subparsers.add_parser("add-match", help="add one (or multiple) match pattern(s) to a place")
    subparser.add_argument("patterns", metavar="PATTERN", nargs="+")
    subparser.set_defaults(func=ClientSession.add_match)

    subparser = subparsers.add_parser("del-match", help="delete one (or multiple) match pattern(s) from a place")
    subparser.add_argument("patterns", metavar="PATTERN", nargs="+")
    subparser.set_defaults(func=ClientSession.del_match)

    subparser = subparsers.add_parser("add-named-match", help="add one match pattern with a name to a place")
    subparser.add_argument("pattern", metavar="PATTERN")
    subparser.add_argument("name", metavar="NAME")
    subparser.set_defaults(func=ClientSession.add_named_match)

    subparser = subparsers.add_parser("acquire", aliases=("lock",), help="acquire a place")
    subparser.add_argument(
        "--allow-unmatched", action="store_true", help="allow missing resources for matches when locking the place"
    )
    subparser.set_defaults(func=ClientSession.acquire)

    subparser = subparsers.add_parser("release", aliases=("unlock",), help="release a place")
    subparser.add_argument(
        "-k", "--kick", action="store_true", help="release a place even if it is acquired by a different user"
    )
    subparser.set_defaults(func=ClientSession.release)

    subparser = subparsers.add_parser(
        "release-from",
        help="atomically release a place, but only if locked by a specific user",
        epilog="Note that this command returns success as long as the specified user no longer owns the place, "
        "meaning it may be acquired by another user or not at all.",
    )
    subparser.add_argument("acquired", metavar="HOST/USER", help="User and host to match against when releasing")
    subparser.set_defaults(func=ClientSession.release_from)

    subparser = subparsers.add_parser("allow", help="allow another user to access a place")
    subparser.add_argument("user", help="<host>/<username>")
    subparser.set_defaults(func=ClientSession.allow)

    subparser = subparsers.add_parser("env", help="generate a labgrid environment file for a place")
    subparser.set_defaults(func=ClientSession.print_env)

    subparser = subparsers.add_parser("power", aliases=("pw",), help="change (or get) a place's power status")
    subparser.add_argument("action", choices=["on", "off", "cycle", "get"])
    subparser.add_argument(
        "-t", "--delay", type=float, default=None, help="wait time in seconds between off and on during cycle"
    )
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.power)

    subparser = subparsers.add_parser("io", help="change (or get) a digital IO status")
    subparser.add_argument("action", choices=["high", "low", "get"], help="action")
    subparser.add_argument("name", help="optional resource name", nargs="?")
    subparser.set_defaults(func=ClientSession.digital_io)

    subparser = subparsers.add_parser("console", aliases=("con",), help="connect to the console")
    subparser.add_argument(
        "-l", "--loop", action="store_true", help="keep trying to connect if the console is unavailable"
    )
    subparser.add_argument(
        "-o", "--listenonly", action="store_true", help="do not modify local terminal, do not send input from stdin"
    )
    subparser.add_argument("name", help="optional resource name", nargs="?")
    subparser.add_argument("--logfile", metavar="FILE", help="Log output to FILE", default=None)
    subparser.set_defaults(func=ClientSession.console)

    subparser = subparsers.add_parser("dfu", help="communicate with device in DFU mode")
    subparser.add_argument("action", choices=["download", "detach", "list"], help="action")
    subparser.add_argument("altsetting", help="altsetting name or number (download, detach only)", nargs="?")
    subparser.add_argument("filename", help="file to write into device (download only)", nargs="?")
    subparser.add_argument("--wait", type=float, default=10.0)
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.dfu)

    subparser = subparsers.add_parser("fastboot", help="run fastboot")
    subparser.add_argument("fastboot_args", metavar="ARG", nargs=argparse.REMAINDER, help="fastboot arguments")
    subparser.add_argument("--wait", type=float, default=10.0)
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.fastboot)

    subparser = subparsers.add_parser("flashscript", help="Run arbitrary script with arguments to flash device")
    subparser.add_argument("script", help="Flashing script")
    subparser.add_argument("script_args", metavar="ARG", nargs=argparse.REMAINDER, help="script arguments")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.flashscript)

    subparser = subparsers.add_parser("bootstrap", help="start a bootloader")
    subparser.add_argument("-w", "--wait", type=float, default=10.0)
    subparser.add_argument("filename", help="filename to boot on the target")
    subparser.add_argument("bootstrap_args", metavar="ARG", nargs=argparse.REMAINDER, help="extra bootstrap arguments")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.bootstrap)

    subparser = subparsers.add_parser("sd-mux", help="switch USB SD Muxer or get current mode")
    subparser.add_argument("action", choices=["dut", "host", "off", "client", "get"])
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.sd_mux)

    subparser = subparsers.add_parser("usb-mux", help="switch USB Muxer")
    subparser.add_argument("links", choices=["off", "dut-device", "host-dut", "host-device", "host-dut+host-device"])
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.usb_mux)

    subparser = subparsers.add_parser(
        "ssh",
        help="connect via ssh (with optional arguments)",
        epilog="Additional arguments are passed to the ssh subprocess.",
    )
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.ssh)

    subparser = subparsers.add_parser("scp", help="transfer file via scp")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument("src", help="source path (use :dir/file for remote side)")
    subparser.add_argument("dst", help="destination path (use :dir/file for remote side)")
    subparser.set_defaults(func=ClientSession.scp)

    subparser = subparsers.add_parser(
        "rsync", help="transfer files via rsync", epilog="Additional arguments are passed to the rsync subprocess."
    )
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument("src", help="source path (use :dir/file for remote side)")
    subparser.add_argument("dst", help="destination path (use :dir/file for remote side)")
    subparser.set_defaults(func=ClientSession.rsync)

    subparser = subparsers.add_parser("sshfs", help="mount via sshfs (blocking)")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument("path", help="remote path on the target")
    subparser.add_argument("mountpoint", help="local path")
    subparser.set_defaults(func=ClientSession.sshfs)

    subparser = subparsers.add_parser("forward", help="forward local port to remote target")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument(
        "--local",
        "-L",
        metavar="[LOCAL:]REMOTE",
        action=LocalPort,
        help="Forward local port LOCAL to remote port REMOTE. If LOCAL is unspecified, an arbitrary port will be chosen",
    )
    subparser.add_argument(
        "--remote",
        "-R",
        metavar="REMOTE:LOCAL",
        action=RemotePort,
        help="Forward remote port REMOTE to local port LOCAL",
    )
    subparser.set_defaults(func=ClientSession.forward)

    subparser = subparsers.add_parser("telnet", help="connect via telnet")
    subparser.set_defaults(func=ClientSession.telnet)

    subparser = subparsers.add_parser("video", help="start a video stream")
    subparser.add_argument("-q", "--quality", type=str, help="select a video quality (use 'list' to show options)")
    subparser.add_argument(
        "-c", "--controls", type=str, help="configure v4l controls (such as 'focus_auto=0,focus_absolute=40')"
    )
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.video)

    subparser = subparsers.add_parser("audio", help="start a audio stream")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.set_defaults(func=ClientSession.audio)

    subparser = subparsers.add_parser("netns", help="start a network namespace with access to NetworkInterface")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument(
        "--address",
        "-a",
        type=ipaddress.ip_interface,
        action="append",
        default=[],
        help="assign address to interface (CIDR notation)",
    )
    subparser.add_argument("--mac-address", help="assign MAC address to interface")
    subparser.add_argument(
        "--nsenter",
        "-e",
        nargs="?",
        default=False,
        const="labgrid-nsenter",
        help="create nsenter wrapper script",
    )
    subparser.add_argument(
        "cmd", nargs=argparse.REMAINDER, help="command to execute in the namespace. If unspecified, $SHELL is invoked"
    )
    subparser.set_defaults(func=ClientSession.netns)

    tmc_parser = subparsers.add_parser("tmc", help="control a USB TMC device")
    tmc_parser.add_argument("--name", "-n", help="optional resource name")
    tmc_parser.set_defaults(func=lambda _: tmc_parser.print_help(file=sys.stderr))
    tmc_subparsers = tmc_parser.add_subparsers(
        dest="subcommand",
        title="available subcommands",
        metavar="SUBCOMMAND",
    )

    tmc_subparser = tmc_subparsers.add_parser("cmd", aliases=("c",), help="execute raw command")
    tmc_subparser.add_argument("command", nargs="+")
    tmc_subparser.set_defaults(func=ClientSession.tmc_command)

    tmc_subparser = tmc_subparsers.add_parser("query", aliases=("q",), help="execute raw query")
    tmc_subparser.add_argument("query", nargs="+")
    tmc_subparser.set_defaults(func=ClientSession.tmc_query)

    tmc_subparser = tmc_subparsers.add_parser("screen", help="show or save a screenshot")
    tmc_subparser.add_argument("action", choices=["show", "save"])
    tmc_subparser.set_defaults(func=ClientSession.tmc_screen)

    tmc_subparser = tmc_subparsers.add_parser("channel", help="use a channel")
    tmc_subparser.add_argument("channel", type=int)
    tmc_subparser.add_argument("action", choices=["info", "values"])
    tmc_subparser.set_defaults(func=ClientSession.tmc_channel)

    subparser = subparsers.add_parser(
        "write-files",
        help="copy files onto mass storage device",
        usage="%(prog)s [OPTION]... -T SOURCE DEST\n" + "       %(prog)s [OPTION]... [-t DIRECTORY] SOURCE...",
    )
    subparser.add_argument("-w", "--wait", type=float, default=10.0, help="storage poll timeout in seconds")
    subparser.add_argument(
        "-p",
        "--partition",
        type=int,
        choices=range(0, 256),
        metavar="0-255",
        default=1,
        help="partition number to mount or 0 to mount whole disk (default: %(default)s)",
    )
    group = subparser.add_mutually_exclusive_group()
    group.add_argument(
        "-t",
        "--target-directory",
        type=pathlib.PurePath,
        metavar="DIRECTORY",
        default=pathlib.PurePath("/"),
        help="copy all SOURCE files into DIRECTORY (default: partition root)",
    )
    group.add_argument("-T", action="store_true", dest="rename", help="copy SOURCE file and rename to DEST")
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument("SOURCE", type=pathlib.PurePath, nargs="+", help="source file(s) to copy")
    subparser.add_argument("DEST", type=pathlib.PurePath, nargs="?", help="destination file name for SOURCE")
    subparser.set_defaults(func=ClientSession.write_files, parser=subparser)

    subparser = subparsers.add_parser("write-image", help="write an image onto mass storage")
    subparser.add_argument("-w", "--wait", type=float, default=10.0)
    subparser.add_argument("-p", "--partition", type=int, help="partition number to write to")
    subparser.add_argument("--skip", type=int, default=0, help="skip n 512-sized blocks at start of input")
    subparser.add_argument("--seek", type=int, default=0, help="skip n 512-sized blocks at start of output")
    subparser.add_argument(
        "--mode",
        dest="write_mode",
        type=Mode,
        choices=Mode,
        default=Mode.DD,
        help="Choose tool for writing images (default: %(default)s)",
    )
    subparser.add_argument("--name", "-n", help="optional resource name")
    subparser.add_argument("filename", help="filename to boot on the target")
    subparser.set_defaults(func=ClientSession.write_image)

    subparser = subparsers.add_parser("reserve", help="create a reservation")
    subparser.add_argument("--wait", action="store_true", help="wait until the reservation is allocated")
    subparser.add_argument("--shell", action="store_true", help="format output as shell variables")
    subparser.add_argument(
        "--prio", type=float, default=0.0, help="priority relative to other reservations (default 0)"
    )
    subparser.add_argument("filters", metavar="KEY=VALUE", nargs="+", help="required tags")
    subparser.set_defaults(func=ClientSession.create_reservation)

    subparser = subparsers.add_parser("cancel-reservation", help="cancel a reservation")
    subparser.add_argument("token", type=str, nargs="?")
    subparser.set_defaults(func=ClientSession.cancel_reservation)

    subparser = subparsers.add_parser("wait", help="wait for a reservation to be allocated")
    subparser.add_argument("token", type=str, nargs="?")
    subparser.set_defaults(func=ClientSession.wait_reservation)

    subparser = subparsers.add_parser("reservations", help="list current reservations")
    subparser.set_defaults(func=ClientSession.print_reservations)

    subparser = subparsers.add_parser(
        "export", help="export driver information to a file (needs environment with drivers)"
    )
    subparser.add_argument(
        "--format",
        dest="format",
        type=ExportFormat,
        choices=ExportFormat,
        default=ExportFormat.SHELL_EXPORT,
        help="output format (default: %(default)s)",
    )
    subparser.add_argument("filename", help="output filename")
    subparser.set_defaults(func=ClientSession.export)

    subparser = subparsers.add_parser("version", help="show version")
    subparser.set_defaults(func=ClientSession.print_version)

    return parser


def main():
    import inspect

    basicConfig(
        level=logging.WARNING,
        stream=sys.stderr,
    )

    StepLogger.start()
    processwrapper.enable_logging()

    # Support both legacy variables and properly namespaced ones
    place = os.environ.get("PLACE", None)
    place = os.environ.get("LG_PLACE", place)
    state = os.environ.get("STATE", None)
    state = os.environ.get("LG_STATE", state)
    initial_state = os.environ.get("LG_INITIAL_STATE", None)
    token = os.environ.get("LG_TOKEN", None)

    parser = get_parser()

    # make any leftover arguments available for some commands
    args, leftover = parser.parse_known_args()
    if args.command not in ["ssh", "rsync", "forward"]:
        args = parser.parse_args()
    else:
        args.leftover = leftover

    # handle dynamic defaults
    if args.config is None:
        args.config = os.environ.get("LG_ENV")

    if args.place is None:
        args.place = place

    if args.state is None:
        args.state = state

    if args.initial_state is None:
        args.initial_state = initial_state

    if args.command in ["cancel-reservation", "wait"] and args.token is None:
        if token:
            args.token = token
        else:
            print("Please provide a token", file=sys.stderr)
            exit(1)

    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)
    if args.verbose > 1:
        logging.getLogger().setLevel(logging.CONSOLE)
    if args.debug or args.verbose > 2:
        logging.getLogger().setLevel(logging.DEBUG)

    if not args.config and (args.state or args.initial_state):
        print("Setting the state requires a configuration file", file=sys.stderr)
        exit(1)
    if args.initial_state and not args.state:
        print("Setting the initial state requires a desired state", file=sys.stderr)
        exit(1)

    if args.proxy:
        proxymanager.force_proxy(args.proxy)

    env = None
    if args.config:
        env = Environment(config_file=args.config)

    role = None
    if args.command != "reserve" and env and env.config.get_targets():
        if args.place:
            if not args.place.startswith("+"):
                role = find_role_by_place(env.config.get_targets(), args.place)
                if not role:
                    print(f"RemotePlace {args.place} not found in configuration file", file=sys.stderr)
                    exit(1)
                print(f"Selected role {role} from configuration file")
        else:
            role, args.place = find_any_role_with_place(env.config.get_targets())
            if not role:
                print("No RemotePlace found in configuration file", file=sys.stderr)
                exit(1)
            print(f"Selected role {role} and place {args.place} from configuration file")

    extra = {
        "args": args,
        "env": env,
        "role": role,
        "prog": parser.prog,
    }

    if args.command and args.command != "help":
        exitcode = 0
        try:
            signal.signal(signal.SIGTERM, lambda *_: sys.exit(0))

            try:
                coordinator_address = args.coordinator or env.config.get_option("coordinator_address")
            except (AttributeError, KeyError):
                # in case of no env or not set, use LG_COORDINATOR env variable or default
                coordinator_address = os.environ.get("LG_COORDINATOR", "127.0.0.1:20408")

            logging.debug('Starting session with "%s"', coordinator_address)
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            session = start_session(coordinator_address, extra=extra, debug=args.debug, loop=loop)
            logging.debug("Started session")

            try:
                if inspect.iscoroutinefunction(args.func):
                    if getattr(args.func, "needs_target", False):
                        place = session.get_acquired_place()
                        target = session._get_target(place)
                        coro = args.func(session, place, target)
                    else:
                        coro = args.func(session)
                    session.loop.run_until_complete(coro)
                else:
                    args.func(session)
            finally:
                logging.debug("Stopping session")
                session.loop.run_until_complete(session.stop())
                session.loop.run_until_complete(session.close())
                logging.debug("Stopping loop")
                session.loop.close()
        except (NoResourceFoundError, NoDriverFoundError, InvalidConfigError) as e:
            if args.debug:
                traceback.print_exc(file=sys.stderr)
            else:
                print(f"{parser.prog}: error: {e}", file=sys.stderr)

            if isinstance(e, NoResourceFoundError):
                if e.found:
                    print("Found multiple resources but no name was given, available names:", file=sys.stderr)
                    for res in e.found:
                        print(f"{res.name}", file=sys.stderr)
                else:
                    print(
                        "This may be caused by disconnected exporter or wrong match entries.\nYou can use the 'show' command to review all matching resources.",
                        file=sys.stderr,
                    )  # pylint: disable=line-too-long
            elif isinstance(e, NoDriverFoundError):
                print(
                    "This is likely caused by an error or missing driver in the environment configuration.",
                    file=sys.stderr,
                )  # pylint: disable=line-too-long
            elif isinstance(e, InvalidConfigError):
                print(
                    "This is likely caused by an error in the environment configuration or invalid\nresource information provided by the coordinator.",
                    file=sys.stderr,
                )  # pylint: disable=line-too-long

            exitcode = 1
        except ServerError as e:
            print(f"Server error: {e}", file=sys.stderr)
            exitcode = 1
        except InteractiveCommandError as e:
            if args.debug:
                traceback.print_exc(file=sys.stderr)
            exitcode = e.exitcode
        except (Error, ErrorGroup) as e:
            if args.debug:
                traceback.print_exc(file=sys.stderr)
            else:
                print(f"{parser.prog}: {e}", file=sys.stderr)
            exitcode = 1
        except KeyboardInterrupt:
            exitcode = 1
        except Exception:  # pylint: disable=broad-except
            traceback.print_exc(file=sys.stderr)
            exitcode = 2
        exit(exitcode)
    else:
        parser.print_help(file=sys.stderr)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labgrid-coordinator.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19labgrid-coordinator.proto\x12\x07labgrid\"\x8a\x01\n\x0f\x43lientInMessage\x12\x1d\n\x04sync\x18\x01 \x01(\x0b\x32\r.labgrid.SyncH\x00\x12\'\n\x07startup\x18\x02 \x01(\x0b\x32\x14.labgrid.StartupDoneH\x00\x12\'\n\tsubscribe\x18\x03 \x01(\x0b\x32\x12.labgrid.SubscribeH\x00\x42\x06\n\x04kind\"\x12\n\x04Sync\x12\n\n\x02id\x18\x01 \x01(\x04\",\n\x0bStartupDone\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"r\n\tSubscribe\x12\x1b\n\x0eis_unsubscribe\x18\x01 \x01(\x08H\x01\x88\x01\x01\x12\x14\n\nall_places\x18\x02 \x01(\x08H\x00\x12\x17\n\rall_resources\x18\x03 \x01(\x08H\x00\x42\x06\n\x04kindB\x11\n\x0f_is_unsubscribe\"g\n\x10\x43lientOutMessage\x12 \n\x04sync\x18\x01 \x01(\x0b\x32\r.labgrid.SyncH\x00\x88\x01\x01\x12(\n\x07updates\x18\x02 \x03(\x0b\x32\x17.labgrid.UpdateResponseB\x07\n\x05_sync\"\xa5\x01\n\x0eUpdateResponse\x12%\n\x08resource\x18\x01 \x01(\x0b\x32\x11.labgrid.ResourceH\x00\x12.\n\x0c\x64\x65l_resource\x18\x02 \x01(\x0b\x32\x16.labgrid.Resource.PathH\x00\x12\x1f\n\x05place\x18\x03 \x01(\x0b\x32\x0e.labgrid.PlaceH\x00\x12\x13\n\tdel_place\x18\x04 \x01(\tH\x00\x42\x06\n\x04kind\"\x9a\x01\n\x11\x45xporterInMessage\x12%\n\x08resource\x18\x01 \x01(\x0b\x32\x11.labgrid.ResourceH\x00\x12\'\n\x07startup\x18\x02 \x01(\x0b\x32\x14.labgrid.StartupDoneH\x00\x12-\n\x08response\x18\x03 \x01(\x0b\x32\x19.labgrid.ExporterResponseH\x00\x42\x06\n\x04kind\"\x9e\x03\n\x08Resource\x12$\n\x04path\x18\x01 \x01(\x0b\x32\x16.labgrid.Resource.Path\x12\x0b\n\x03\x63ls\x18\x02 \x01(\t\x12-\n\x06params\x18\x03 \x03(\x0b\x32\x1d.labgrid.Resource.ParamsEntry\x12+\n\x05\x65xtra\x18\x04 \x03(\x0b\x32\x1c.labgrid.Resource.ExtraEntry\x12\x10\n\x08\x61\x63quired\x18\x05 \x01(\t\x12\r\n\x05\x61vail\x18\x06 \x01(\x08\x1a_\n\x04Path\x12\x1a\n\rexporter_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\ngroup_name\x18\x02 \x01(\t\x12\x15\n\rresource_name\x18\x03 \x01(\tB\x10\n\x0e_exporter_name\x1a@\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.labgrid.MapValue:\x02\x38\x01\x1a?\n\nExtraEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.labgrid.MapValue:\x02\x38\x01\"\x82\x01\n\x08MapValue\x12\x14\n\nbool_value\x18\x01 \x01(\x08H\x00\x12\x13\n\tint_value\x18\x02 \x01(\x03H\x00\x12\x14\n\nuint_value\x18\x03 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x04 \x01(\x01H\x00\x12\x16\n\x0cstring_value\x18\x05 \x01(\tH\x00\x42\x06\n\x04kind\"C\n\x10\x45xporterResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x13\n\x06reason\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_reason\"\x18\n\x05Hello\x12\x0f\n\x07version\x18\x01 \x01(\t\"\x82\x01\n\x12\x45xporterOutMessage\x12\x1f\n\x05hello\x18\x01 \x01(\x0b\x32\x0e.labgrid.HelloH\x00\x12\x43\n\x14set_acquired_request\x18\x02 \x01(\x0b\x32#.labgrid.ExporterSetAcquiredRequestH\x00\x42\x06\n\x04kind\"o\n\x1a\x45xporterSetAcquiredRequest\x12\x12\n\ngroup_name\x18\x01 \x01(\t\x12\x15\n\rresource_name\x18\x02 \x01(\t\x12\x17\n\nplace_name\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\r\n\x0b_place_name\"\x1f\n\x0f\x41\x64\x64PlaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x12\n\x10\x41\x64\x64PlaceResponse\"\"\n\x12\x44\x65letePlaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x15\n\x13\x44\x65letePlaceResponse\"\x12\n\x10GetPlacesRequest\"3\n\x11GetPlacesResponse\x12\x1e\n\x06places\x18\x01 \x03(\x0b\x32\x0e.labgrid.Place\"\xd2\x02\n\x05Place\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x61liases\x18\x02 \x03(\t\x12\x0f\n\x07\x63omment\x18\x03 \x01(\t\x12&\n\x04tags\x18\x04 \x03(\x0b\x32\x18.labgrid.Place.TagsEntry\x12\'\n\x07matches\x18\x05 \x03(\x0b\x32\x16.labgrid.ResourceMatch\x12\x15\n\x08\x61\x63quired\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x12\x61\x63quired_resources\x18\x07 \x03(\t\x12\x0f\n\x07\x61llowed\x18\x08 \x03(\t\x12\x0f\n\x07\x63reated\x18\t \x01(\x01\x12\x0f\n\x07\x63hanged\x18\n \x01(\x01\x12\x18\n\x0breservation\x18\x0b \x01(\tH\x01\x88\x01\x01\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0b\n\t_acquiredB\x0e\n\x0c_reservation\"y\n\rResourceMatch\x12\x10\n\x08\x65xporter\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0b\n\x03\x63ls\x18\x03 \x01(\t\x12\x11\n\x04name\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06rename\x18\x05 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_nameB\t\n\x07_rename\"8\n\x14\x41\x64\x64PlaceAliasRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"\x17\n\x15\x41\x64\x64PlaceAliasResponse\";\n\x17\x44\x65letePlaceAliasRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"\x1a\n\x18\x44\x65letePlaceAliasResponse\"\x8b\x01\n\x13SetPlaceTagsRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x34\n\x04tags\x18\x02 \x03(\x0b\x32&.labgrid.SetPlaceTagsRequest.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x16\n\x14SetPlaceTagsResponse\"<\n\x16SetPlaceCommentRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\"\x19\n\x17SetPlaceCommentResponse\"Z\n\x14\x41\x64\x64PlaceMatchRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\x13\n\x06rename\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_rename\"\x17\n\x15\x41\x64\x64PlaceMatchResponse\"]\n\x17\x44\x65letePlaceMatchRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\x13\n\x06rename\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_rename\"\x1a\n\x18\x44\x65letePlaceMatchResponse\"(\n\x13\x41\x63quirePlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\"\x16\n\x14\x41\x63quirePlaceResponse\"L\n\x13ReleasePlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x15\n\x08\x66romuser\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x0b\n\t_fromuser\"\x16\n\x14ReleasePlaceResponse\"4\n\x11\x41llowPlaceRequest\x12\x11\n\tplacename\x18\x01 \x01(\t\x12\x0c\n\x04user\x18\x02 \x01(\t\"\x14\n\x12\x41llowPlaceResponse\"\xb6\x01\n\x18\x43reateReservationRequest\x12?\n\x07\x66ilters\x18\x01 \x03(\x0b\x32..labgrid.CreateReservationRequest.FiltersEntry\x12\x0c\n\x04prio\x18\x02 \x01(\x01\x1aK\n\x0c\x46iltersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.labgrid.Reservation.Filter:\x02\x38\x01\"F\n\x19\x43reateReservationResponse\x12)\n\x0breservation\x18\x01 \x01(\x0b\x32\x14.labgrid.Reservation\"\xcd\x03\n\x0bReservation\x12\r\n\x05owner\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\x05\x12\x0c\n\x04prio\x18\x04 \x01(\x01\x12\x32\n\x07\x66ilters\x18\x05 \x03(\x0b\x32!.labgrid.Reservation.FiltersEntry\x12:\n\x0b\x61llocations\x18\x06 \x03(\x0b\x32%.labgrid.Reservation.AllocationsEntry\x12\x0f\n\x07\x63reated\x18\x07 \x01(\x01\x12\x0f\n\x07timeout\x18\x08 \x01(\x01\x1ap\n\x06\x46ilter\x12\x37\n\x06\x66ilter\x18\x01 \x03(\x0b\x32\'.labgrid.Reservation.Filter.FilterEntry\x1a-\n\x0b\x46ilterEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aK\n\x0c\x46iltersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.labgrid.Reservation.Filter:\x02\x38\x01\x1a\x32\n\x10\x41llocationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\")\n\x18\x43\x61ncelReservationRequest\x12\r\n\x05token\x18\x01 \x01(\t\"\x1b\n\x19\x43\x61ncelReservationResponse\"\'\n\x16PollReservationRequest\x12\r\n\x05token\x18\x01 \x01(\t\"D\n\x17PollReservationResponse\x12)\n\x0breservation\x18\x01 \x01(\x0b\x32\x14.labgrid.Reservation\"E\n\x17GetReservationsResponse\x12*\n\x0creservations\x18\x01 \x03(\x0b\x32\x14.labgrid.Reservation\"\x18\n\x16GetReservationsRequest2\xd2\x0b\n\x0b\x43oordinator\x12I\n\x0c\x43lientStream\x12\x18.labgrid.ClientInMessage\x1a\x19.labgrid.ClientOutMessage\"\x00(\x01\x30\x01\x12O\n\x0e\x45xporterStream\x12\x1a.labgrid.ExporterInMessage\x1a\x1b.labgrid.ExporterOutMessage\"\x00(\x01\x30\x01\x12\x41\n\x08\x41\x64\x64Place\x12\x18.labgrid.AddPlaceRequest\x1a\x19.labgrid.AddPlaceResponse\"\x00\x12J\n\x0b\x44\x65letePlace\x12\x1b.labgrid.DeletePlaceRequest\x1a\x1c.labgrid.DeletePlaceResponse\"\x00\x12\x44\n\tGetPlaces\x12\x19.labgrid.GetPlacesRequest\x1a\x1a.labgrid.GetPlacesResponse\"\x00\x12P\n\rAddPlaceAlias\x12\x1d.labgrid.AddPlaceAliasRequest\x1a\x1e.labgrid.AddPlaceAliasResponse\"\x00\x12Y\n\x10\x44\x65letePlaceAlias\x12 .labgrid.DeletePlaceAliasRequest\x1a!.labgrid.DeletePlaceAliasResponse\"\x00\x12M\n\x0cSetPlaceTags\x12\x1c.labgrid.SetPlaceTagsRequest\x1a\x1d.labgrid.SetPlaceTagsResponse\"\x00\x12V\n\x0fSetPlaceComment\x12\x1f.labgrid.SetPlaceCommentRequest\x1a .labgrid.SetPlaceCommentResponse\"\x00\x12P\n\rAddPlaceMatch\x12\x1d.labgrid.AddPlaceMatchRequest\x1a\x1e.labgrid.AddPlaceMatchResponse\"\x00\x12Y\n\x10\x44\x65letePlaceMatch\x12 .labgrid.DeletePlaceMatchRequest\x1a!.labgrid.DeletePlaceMatchResponse\"\x00\x12M\n\x0c\x41\x63quirePlace\x12\x1c.labgrid.AcquirePlaceRequest\x1a\x1d.labgrid.AcquirePlaceResponse\"\x00\x12M\n\x0cReleasePlace\x12\x1c.labgrid.ReleasePlaceRequest\x1a\x1d.labgrid.ReleasePlaceResponse\"\x00\x12G\n\nAllowPlace\x12\x1a.labgrid.AllowPlaceRequest\x1a\x1b.labgrid.AllowPlaceResponse\"\x00\x12\\\n\x11\x43reateReservation\x12!.labgrid.CreateReservationRequest\x1a\".labgrid.CreateReservationResponse\"\x00\x12\\\n\x11\x43\x61ncelReservation\x12!.labgrid.CancelReservationRequest\x1a\".labgrid.CancelReservationResponse\"\x00\x12V\n\x0fPollReservation\x12\x1f.labgrid.PollReservationRequest\x1a .labgrid.PollReservationResponse\"\x00\x12V\n\x0fGetReservations\x12\x1f.labgrid.GetReservationsRequest\x1a .labgrid.GetReservationsResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'labgrid_coordinator_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_RESOURCE_PARAMSENTRY']._options = None
  _globals['_RESOURCE_PARAMSENTRY']._serialized_options = b'8\001'
  _globals['_RESOURCE_EXTRAENTRY']._options = None
  _globals['_RESOURCE_EXTRAENTRY']._serialized_options = b'8\001'
  _globals['_PLACE_TAGSENTRY']._options = None
  _globals['_PLACE_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._options = None
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._options = None
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_FILTER_FILTERENTRY']._options = None
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_FILTERSENTRY']._options = None
  _globals['_RESERVATION_FILTERSENTRY']._serialized_options = b'8\001'
  _globals['_RESERVATION_ALLOCATIONSENTRY']._options = None
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_options = b'8\001'
  _globals['_CLIENTINMESSAGE']._serialized_start=39
  _globals['_CLIENTINMESSAGE']._serialized_end=177
  _globals['_SYNC']._serialized_start=179
  _globals['_SYNC']._serialized_end=197
  _globals['_STARTUPDONE']._serialized_start=199
  _globals['_STARTUPDONE']._serialized_end=243
  _globals['_SUBSCRIBE']._serialized_start=245
  _globals['_SUBSCRIBE']._serialized_end=359
  _globals['_CLIENTOUTMESSAGE']._serialized_start=361
  _globals['_CLIENTOUTMESSAGE']._serialized_end=464
  _globals['_UPDATERESPONSE']._serialized_start=467
  _globals['_UPDATERESPONSE']._serialized_end=632
  _globals['_EXPORTERINMESSAGE']._serialized_start=635
  _globals['_EXPORTERINMESSAGE']._serialized_end=789
  _globals['_RESOURCE']._serialized_start=792
  _globals['_RESOURCE']._serialized_end=1206
  _globals['_RESOURCE_PATH']._serialized_start=980
  _globals['_RESOURCE_PATH']._serialized_end=1075
  _globals['_RESOURCE_PARAMSENTRY']._serialized_start=1077
  _globals['_RESOURCE_PARAMSENTRY']._serialized_end=1141
  _globals['_RESOURCE_EXTRAENTRY']._serialized_start=1143
  _globals['_RESOURCE_EXTRAENTRY']._serialized_end=1206
  _globals['_MAPVALUE']._serialized_start=1209
  _globals['_MAPVALUE']._serialized_end=1339
  _globals['_EXPORTERRESPONSE']._serialized_start=1341
  _globals['_EXPORTERRESPONSE']._serialized_end=1408
  _globals['_HELLO']._serialized_start=1410
  _globals['_HELLO']._serialized_end=1434
  _globals['_EXPORTEROUTMESSAGE']._serialized_start=1437
  _globals['_EXPORTEROUTMESSAGE']._serialized_end=1567
  _globals['_EXPORTERSETACQUIREDREQUEST']._serialized_start=1569
  _globals['_EXPORTERSETACQUIREDREQUEST']._serialized_end=1680
  _globals['_ADDPLACEREQUEST']._serialized_start=1682
  _globals['_ADDPLACEREQUEST']._serialized_end=1713
  _globals['_ADDPLACERESPONSE']._serialized_start=1715
  _globals['_ADDPLACERESPONSE']._serialized_end=1733
  _globals['_DELETEPLACEREQUEST']._serialized_start=1735
  _globals['_DELETEPLACEREQUEST']._serialized_end=1769
  _globals['_DELETEPLACERESPONSE']._serialized_start=1771
  _globals['_DELETEPLACERESPONSE']._serialized_end=1792
  _globals['_GETPLACESREQUEST']._serialized_start=1794
  _globals['_GETPLACESREQUEST']._serialized_end=1812
  _globals['_GETPLACESRESPONSE']._serialized_start=1814
  _globals['_GETPLACESRESPONSE']._serialized_end=1865
  _globals['_PLACE']._serialized_start=1868
  _globals['_PLACE']._serialized_end=2206
  _globals['_PLACE_TAGSENTRY']._serialized_start=2134
  _globals['_PLACE_TAGSENTRY']._serialized_end=2177
  _globals['_RESOURCEMATCH']._serialized_start=2208
  _globals['_RESOURCEMATCH']._serialized_end=2329
  _globals['_ADDPLACEALIASREQUEST']._serialized_start=2331
  _globals['_ADDPLACEALIASREQUEST']._serialized_end=2387
  _globals['_ADDPLACEALIASRESPONSE']._serialized_start=2389
  _globals['_ADDPLACEALIASRESPONSE']._serialized_end=2412
  _globals['_DELETEPLACEALIASREQUEST']._serialized_start=2414
  _globals['_DELETEPLACEALIASREQUEST']._serialized_end=2473
  _globals['_DELETEPLACEALIASRESPONSE']._serialized_start=2475
  _globals['_DELETEPLACEALIASRESPONSE']._serialized_end=2501
  _globals['_SETPLACETAGSREQUEST']._serialized_start=2504
  _globals['_SETPLACETAGSREQUEST']._serialized_end=2643
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_start=2134
  _globals['_SETPLACETAGSREQUEST_TAGSENTRY']._serialized_end=2177
  _globals['_SETPLACETAGSRESPONSE']._serialized_start=2645
  _globals['_SETPLACETAGSRESPONSE']._serialized_end=2667
  _globals['_SETPLACECOMMENTREQUEST']._serialized_start=2669
  _globals['_SETPLACECOMMENTREQUEST']._serialized_end=2729
  _globals['_SETPLACECOMMENTRESPONSE']._serialized_start=2731
  _globals['_SETPLACECOMMENTRESPONSE']._serialized_end=2756
  _globals['_ADDPLACEMATCHREQUEST']._serialized_start=2758
  _globals['_ADDPLACEMATCHREQUEST']._serialized_end=2848
  _globals['_ADDPLACEMATCHRESPONSE']._serialized_start=2850
  _globals['_ADDPLACEMATCHRESPONSE']._serialized_end=2873
  _globals['_DELETEPLACEMATCHREQUEST']._serialized_start=2875
  _globals['_DELETEPLACEMATCHREQUEST']._serialized_end=2968
  _globals['_DELETEPLACEMATCHRESPONSE']._serialized_start=2970
  _globals['_DELETEPLACEMATCHRESPONSE']._serialized_end=2996
  _globals['_ACQUIREPLACEREQUEST']._serialized_start=2998
  _globals['_ACQUIREPLACEREQUEST']._serialized_end=3038
  _globals['_ACQUIREPLACERESPONSE']._serialized_start=3040
  _globals['_ACQUIREPLACERESPONSE']._serialized_end=3062
  _globals['_RELEASEPLACEREQUEST']._serialized_start=3064
  _globals['_RELEASEPLACEREQUEST']._serialized_end=3140
  _globals['_RELEASEPLACERESPONSE']._serialized_start=3142
  _globals['_RELEASEPLACERESPONSE']._serialized_end=3164
  _globals['_ALLOWPLACEREQUEST']._serialized_start=3166
  _globals['_ALLOWPLACEREQUEST']._serialized_end=3218
  _globals['_ALLOWPLACERESPONSE']._serialized_start=3220
  _globals['_ALLOWPLACERESPONSE']._serialized_end=3240
  _globals['_CREATERESERVATIONREQUEST']._serialized_start=3243
  _globals['_CREATERESERVATIONREQUEST']._serialized_end=3425
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_start=3350
  _globals['_CREATERESERVATIONREQUEST_FILTERSENTRY']._serialized_end=3425
  _globals['_CREATERESERVATIONRESPONSE']._serialized_start=3427
  _globals['_CREATERESERVATIONRESPONSE']._serialized_end=3497
  _globals['_RESERVATION']._serialized_start=3500
  _globals['_RESERVATION']._serialized_end=3961
  _globals['_RESERVATION_FILTER']._serialized_start=3720
  _globals['_RESERVATION_FILTER']._serialized_end=3832
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_start=3787
  _globals['_RESERVATION_FILTER_FILTERENTRY']._serialized_end=3832
  _globals['_RESERVATION_FILTERSENTRY']._serialized_start=3350
  _globals['_RESERVATION_FILTERSENTRY']._serialized_end=3425
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_start=3911
  _globals['_RESERVATION_ALLOCATIONSENTRY']._serialized_end=3961
  _globals['_CANCELRESERVATIONREQUEST']._serialized_start=3963
  _globals['_CANCELRESERVATIONREQUEST']._serialized_end=4004
  _globals['_CANCELRESERVATIONRESPONSE']._serialized_start=4006
  _globals['_CANCELRESERVATIONRESPONSE']._serialized_end=4033
  _globals['_POLLRESERVATIONREQUEST']._serialized_start=4035
  _globals['_POLLRESERVATIONREQUEST']._serialized_end=4074
  _globals['_POLLRESERVATIONRESPONSE']._serialized_start=4076
  _globals['_POLLRESERVATIONRESPONSE']._serialized_end=4144
  _globals['_GETRESERVATIONSRESPONSE']._serialized_start=4146
  _globals['_GETRESERVATIONSRESPONSE']._serialized_end=4215
  _globals['_GETRESERVATIONSREQUEST']._serialized_start=4217
  _globals['_GETRESERVATIONSREQUEST']._serialized_end=4241
  _globals['_COORDINATOR']._serialized_start=4244
  _globals['_COORDINATOR']._serialized_end=5734
# @@protoc_insertion_point(module_scope)



import enum

import attr

from ..factory import target_factory
from .common import Strategy, StrategyError, never_retry


class Status(enum.Enum):
    unknown = 0
    off = 1
    uboot = 2
    shell = 3


@target_factory.reg_driver
@attr.s(eq=False)
class UBootStrategy(Strategy):
    """UbootStrategy - Strategy to switch to uboot or shell"""
    bindings = {
        "power": "PowerProtocol",
        "console": "ConsoleProtocol",
        "uboot": "UBootDriver",
        "shell": "ShellDriver",
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    def transition(self, status):
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.unknown:
            raise StrategyError(f"can not transition to {status}")
        elif status == self.status:
            return # nothing to do
        elif status == Status.off:
            self.target.deactivate(self.console)
            self.target.activate(self.power)
            self.power.off()
        elif status == Status.uboot:
            self.transition(Status.off)
            self.target.activate(self.console)
            # cycle power
            self.power.cycle()
            # interrupt uboot
            self.target.activate(self.uboot)
        elif status == Status.shell:
            # transition to uboot
            self.transition(Status.uboot)
            self.uboot.boot("")
            self.uboot.await_boot()
            self.target.activate(self.shell)
            self.shell.run("systemctl is-system-running --wait")
        else:
            raise StrategyError(f"no transition found from {self.status} to {status}")
        self.status = status

    @never_retry
    def force(self, status):
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.off:
            self.target.activate(self.power)
        elif status == Status.uboot:
            self.target.activate(self.uboot)
        elif status == Status.shell:
            self.target.activate(self.shell)
        else:
            raise StrategyError("can not force state {}".format(status))
        self.status = status



"""The U-Boot Module contains the UBootDriver"""
import re
import attr
from pexpect import TIMEOUT

from ..factory import target_factory
from ..protocol import CommandProtocol, ConsoleProtocol, LinuxBootProtocol
from ..util import gen_marker, re_vt100, Timeout
from ..step import step
from .common import Driver
from .commandmixin import CommandMixin

re_uboot_timestamp = re.compile(r"^\[\s+\d+\.\d+\]\s*")

@target_factory.reg_driver
@attr.s(eq=False)
class UBootDriver(CommandMixin, Driver, CommandProtocol, LinuxBootProtocol):
    """UBootDriver - Driver to control uboot via the console.
    UBootDriver binds on top of a ConsoleProtocol.

    Args:
        prompt (str): optional, U-Boot prompt
        password (str): optional, password to unlock U-Boot
        init_commands (tuple): optional, tuple of commands to run after unlock
        autoboot (str): optional, string to search for to interrupt autoboot
        interrupt (str): optional, character to interrupt autoboot and go to prompt
        password_prompt (str): optional, string to detect the password prompt
        boot_expression (str): optional, deprecated
        bootstring (str): optional, string that indicates that the Kernel is booting
        boot_command (str): optional boot command to boot target
        login_timeout (int): optional, timeout for login prompt detection
        boot_timeout (int): optional, timeout for initial Linux Kernel version detection
        strip_timestamp (bool, default=False): strip timestamps prepended to console when present

    """
    bindings = {"console": ConsoleProtocol, }
    prompt = attr.ib(default="", validator=attr.validators.instance_of(str))
    autoboot = attr.ib(default="stop autoboot", validator=attr.validators.instance_of(str))
    password = attr.ib(default="", validator=attr.validators.instance_of(str))
    interrupt = attr.ib(default="\n", validator=attr.validators.instance_of(str))
    init_commands = attr.ib(default=attr.Factory(tuple), converter=tuple)
    password_prompt = attr.ib(default="enter Password:", validator=attr.validators.instance_of(str))
    boot_expression = attr.ib(default="", validator=attr.validators.instance_of(str))
    bootstring = attr.ib(default=r"Linux version \d", validator=attr.validators.instance_of(str))
    boot_command = attr.ib(default="run bootcmd", validator=attr.validators.instance_of(str))
    boot_commands = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))
    login_timeout = attr.ib(default=30, validator=attr.validators.instance_of(int))
    boot_timeout = attr.ib(default=30, validator=attr.validators.instance_of(int))
    strip_timestamp = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._status = 0

        if self.boot_expression:
            import warnings
            warnings.warn("boot_expression is deprecated and will be ignored", DeprecationWarning)

    def on_activate(self):
        """Activate the UBootDriver

        This function checks for a prompt and awaits it if not already active
        """
        if self._status == 0:
            self._await_prompt()

    def on_deactivate(self):
        """Deactivate the UBootDriver

        Simply sets the internal status to 0
        """
        self._status = 0

    def _run(self, cmd: str, *, timeout: int = 30, codec: str = "utf-8", decodeerrors: str = "strict"):  # pylint: disable=unused-argument,line-too-long
        # TODO: use codec, decodeerrors
        # TODO: Shell Escaping for the U-Boot Shell
        marker = gen_marker()
        cmp_command = f"""echo '{marker[:4]}''{marker[4:]}'; {cmd}; echo "$?"; echo '{marker[:4]}''{marker[4:]}';"""  # pylint: disable=line-too-long
        if self._status == 1:
            self.console.sendline(cmp_command)
            _, before, _, _ = self.console.expect(self.prompt, timeout=timeout)
            # Remove VT100 Codes and split by newline
            data = re_vt100.sub(
                '', before.decode('utf-8'), count=1000000
            ).replace("\r", "").split("\n")

            # Strip possible U-Boot timestamps from the line
            if self.strip_timestamp:
                data = [re_uboot_timestamp.sub('', line) for line in data]

            self.logger.debug("Received Data: %s", data)
            # Remove first element, the invoked cmd
            data = data[data.index(marker) + 1:]
            data = data[:data.index(marker)]
            exitcode = int(data[-1])
            del data[-1]
            return (data, [], exitcode)

        return None

    @Driver.check_active
    @step(args=['cmd'], result=True)
    def run(self, cmd, timeout=30):
        """
        Runs the specified command on the shell and returns the output.

        Args:
            cmd (str): command to run on the shell
            timeout (int): optional, how long to wait for completion

        Returns:
            Tuple[List[str],List[str], int]: if successful, None otherwise
        """
        return self._run(cmd, timeout=timeout)

    def get_status(self):
        """Retrieve status of the UBootDriver.
        0 means inactive, 1 means active.

        Returns:
            int: status of the driver
        """
        return self._status

    def _check_prompt(self):
        """
        Internal function to check if we have a valid prompt.
        It sets the internal _status to 1 or 0 based on the prompt detection.
        """
        marker = gen_marker()
        # hide marker from expect
        self.console.sendline(f"echo '{marker[:4]}''{marker[4:]}'")
        try:
            self.console.expect(f"{marker}", timeout=2)
            self.console.expect(self.prompt, timeout=1)
            self._status = 1
        except TIMEOUT:
            self._status = 0
            raise

    @Driver.check_active
    @step()
    def reset(self):
        """Reset the board via a CPU reset
        """
        self._status = 0
        self.console.sendline("reset")
        self._await_prompt()

    @step()
    def _await_prompt(self):
        """Await autoboot line and stop it to get to the prompt, optionally
        enter the password.
        """
        timeout = Timeout(float(self.login_timeout))

        # We call console.expect with a short timeout here to detect if the
        # console is idle, which would result in a timeout without any changes
        # to the before property. So we store the last before value we've seen.
        # Because pexpect keeps any read data in it's buffer when a timeout
        # occours, we can't lose any data this way.
        last_before = None

        expectations = [self.prompt, self.autoboot, self.password_prompt, TIMEOUT]
        while True:
            index, before, _, _ = self.console.expect(
                expectations,
                timeout=2
            )
            if index == 0:
                self._status = 1
                break

            elif index == 1:
                self.console.write(self.interrupt.encode('ASCII'))

            elif index == 2:
                if not self.password:
                    raise Exception("Password entry needed but no password set")
                self.console.sendline(self.password)

            elif index == 3:
                # expect hit a timeout while waiting for a match
                if before == last_before:
                    # we did not receive anything during the previous expect cycle
                    # let's assume the target is idle and we can safely issue a
                    # newline to check the state
                    self.console.sendline("")

                if timeout.expired:
                    raise TIMEOUT(
                        f"Timeout of {self.login_timeout} seconds exceeded during waiting for login"
                    )

            last_before = before

        if self.prompt:
            self._check_prompt()

        for command in self.init_commands:
            self._run_check(command)

    @Driver.check_active
    @step()
    def await_boot(self):
        """Wait for the initial Linux version string to verify we successfully
        jumped into the kernel.
        """
        self.console.expect(self.bootstring, timeout=self.boot_timeout)

    @Driver.check_active
    @step(args=['name'])
    def boot(self, name: str = ""):
        """Boot the default or a specific boot entry

        Args:
            name (str): name of the entry to boot"""
        if name:
            try:
                self.console.sendline(self.boot_commands[name])
            except KeyError as e:
                raise Exception(f"{name} not found in boot_commands") from e
        else:
            self.console.sendline(self.boot_command)



"""
This module contains helper functions for working with dictionaries.
"""

import warnings

import attr


def diff_dict(old, new):
    """
    Compares old and new dictionaries, yielding for each difference (key,
    old_value, new_value).
    None is used for missing values.
    """
    for key in sorted(old.keys() | new.keys()):
        v_old = old.get(key)
        v_new = new.get(key)
        if v_old != v_new:
            yield key, v_old, v_new


def flat_dict(d):
    def flatten(d, prefix=()):
        for key, value in d.items():
            key = prefix + (key,)
            if isinstance(value, dict):
                yield from flatten(value, prefix=key)
            else:
                yield ".".join(key), value

    return dict(flatten(d))


def filter_dict(d, cls, warn=False):
    """
    Returns a copy a dictionary which only contains the attributes defined on
    an attrs class.
    """
    assert attr.has(cls)
    fields = set(a.name for a in attr.fields(cls))
    if warn:
        remove = set(d) - fields
        for k in sorted(remove):
            warnings.warn(f"unsupported attribute '{k}' with value '{d[k]}' for class '{cls.__name__}'", stacklevel=2)
    return {k: v for k, v in d.items() if k in fields}


def find_dict(d, key):
    """
    Recursively search for a key in a dictionary

    Args:
        d (dict): The dictionary to recursively search through
        key (str): The key to search for
    """
    if key in d:
        return d[key]
    for v in d.values():
        if isinstance(v, dict):
            value = find_dict(v, key)
            if value is not None:
                return value



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class LAASerialPort(Resource):
    """This resource describes a serial port on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection
        serial_name (str): name of the serial port on the LAA"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))
    serial_name = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class LAAPowerPort(Resource):
    """This resource describes a power port on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection
        power_on (list): sequence of (vbus, state) tuples for power on
        power_off (list): sequence of (vbus, state) tuples for power off
        power_cycle (list): optional sequence for power cycle"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))
    power_on = attr.ib(validator=attr.validators.instance_of(list))
    power_off = attr.ib(validator=attr.validators.instance_of(list))
    power_cycle = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(list)),
    )
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        for name in ("power_on", "power_off"):
            self._check_power_sequence(name, getattr(self, name))
        if self.power_cycle is not None:
            self._check_power_sequence("power_cycle", self.power_cycle)

    @staticmethod
    def _check_power_sequence(name, seq):
        for i, entry in enumerate(seq):
            if not isinstance(entry, (list, tuple)) or len(entry) != 2:
                raise ValueError(
                    f"{name}[{i}] must be [vbus, state], got {entry!r}"
                )
            if not isinstance(entry[0], str) or not isinstance(entry[1], str):
                raise ValueError(
                    f"{name}[{i}] must be [str, str], got {entry!r}"
                )


@target_factory.reg_resource
@attr.s(eq=False)
class LAAUSBGadgetMassStorage(Resource):
    """This resource describes a USB gadget mass storage device on a LAA.

    Args:
        laa_identity (str): LAA identity for connection
        image (str): mass storage image filename"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))
    image = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class LAAUSBPort(Resource):
    """This resource describes USB ports on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection
        usb_ports (list): list of USB port numbers to control"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))
    usb_ports = attr.ib(validator=attr.validators.instance_of(list))


@target_factory.reg_resource
@attr.s(eq=False)
class LAAButtonPort(Resource):
    """This resource describes virtual buttons on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection
        buttons (list): list of button names available"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))
    buttons = attr.ib(validator=attr.validators.instance_of(list))


@target_factory.reg_resource
@attr.s(eq=False)
class LAALed(Resource):
    """This resource describes a LED on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class LAATempSensor(Resource):
    """This resource describes a temperature sensor on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class LAAWattMeter(Resource):
    """This resource describes a power meter on a DUT connected via LAA.

    Args:
        laa_identity (str): LAA identity for connection"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_resource
@attr.s(eq=False)
class LAAProvider(Resource):
    """This resource describes file storage on an LAA for TFTP provisioning.

    Args:
        laa_identity (str): LAA identity for connection"""
    laa_identity = attr.ib(validator=attr.validators.instance_of(str))





import subprocess
import attr

from ..factory import target_factory
from ..protocol import BootstrapProtocol
from ..step import step
from .common import Driver
from ..util.managedfile import ManagedFile
from ..util.timeout import Timeout
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class MXSUSBDriver(Driver, BootstrapProtocol):
    bindings = {
        "loader": {"MXSUSBLoader", "NetworkMXSUSBLoader"},
    }

    image = attr.ib(default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('mxs-usb-loader')
        else:
            self.tool = 'mxs-usb-loader'

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=['filename'])
    def load(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.loader)
        mf.sync_to_resource()

        processwrapper.check_output(
            self.loader.command_prefix + [self.tool, "0", mf.get_remote_path()],
            print_on_silent_log=True
        )


@target_factory.reg_driver
@attr.s(eq=False)
class IMXUSBDriver(Driver, BootstrapProtocol):
    bindings = {
        "loader": {"IMXUSBLoader", "NetworkIMXUSBLoader", "MXSUSBLoader", "NetworkMXSUSBLoader"},
    }

    image = attr.ib(default=None)
    verify = attr.ib(default=True, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('imx-usb-loader')
        else:
            self.tool = 'imx-usb-loader'

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=['filename'])
    def load(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.loader)
        mf.sync_to_resource()

        command = [self.tool, "-p", str(self.loader.path)]
        if self.verify:
            command.append("-c")
        command.append(mf.get_remote_path())

        processwrapper.check_output(
            self.loader.command_prefix + command,
            print_on_silent_log=True
        )


@target_factory.reg_driver
@attr.s(eq=False)
class RKUSBDriver(Driver, BootstrapProtocol):
    bindings = {
        "loader": {"RKUSBLoader", "NetworkRKUSBLoader"},
    }

    image = attr.ib(default=None)
    usb_loader = attr.ib(default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('rk-usb-loader')
        else:
            self.tool = 'rk-usb-loader'

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=['filename'])
    def load(self, filename=None):
        if self.target.env:
            usb_loader = self.target.env.config.get_image_path(self.usb_loader)
            mf = ManagedFile(usb_loader, self.loader)
            mf.sync_to_resource()

        timeout = Timeout(3.0)
        while True:
            try:
                processwrapper.check_output(
                    self.loader.command_prefix +
                    [self.tool, 'db', mf.get_remote_path()],
                    print_on_silent_log=True
                )
                break
            except subprocess.CalledProcessError:
                if timeout.expired:
                    raise

        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.loader)
        mf.sync_to_resource()

        timeout = Timeout(3.0)
        while True:
            try:
                processwrapper.check_output(
                    self.loader.command_prefix +
                    [self.tool, 'wl', '0x40', mf.get_remote_path()],
                    print_on_silent_log=True
                )
                break
            except subprocess.CalledProcessError:
                if timeout.expired:
                    raise


@target_factory.reg_driver
@attr.s(eq=False)
class UUUDriver(Driver, BootstrapProtocol):
    bindings = {
        "loader": {"IMXUSBLoader", "NetworkIMXUSBLoader", "MXSUSBLoader", "NetworkMXSUSBLoader"},
    }

    image = attr.ib(default=None)
    script = attr.ib(default='', validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('uuu-loader')
        else:
            self.tool = 'uuu-loader'

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=['filename'])
    def load(self, filename=None):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.loader)
        mf.sync_to_resource()

        cmd = ['-b', self.script] if self.script else []

        processwrapper.check_output(
            self.loader.command_prefix + [self.tool] + cmd + [mf.get_remote_path()],
            print_on_silent_log=True
        )


@target_factory.reg_driver
@attr.s(eq=False)
class BDIMXUSBDriver(Driver, BootstrapProtocol):
    """
    This is a Driver for the Boundary Devices imx_usb_loader available from
    https://github.com/boundarydevices/imx_usb_loader

    It supports loading the second stage bootloader to a SDP gadget implemented
    by the first stage bootloader in SRAM. Accordingly, the image to upload
    must be specified explicitly.
    """
    bindings = {
        "loader": {"IMXUSBLoader", "NetworkIMXUSBLoader"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool('imx_usb')
        else:
            self.tool = 'imx_usb'

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=['filename'])
    def load(self, filename):
        mf = ManagedFile(filename, self.loader)
        mf.sync_to_resource()

        processwrapper.check_output(
            self.loader.command_prefix + [
                self.tool,
                f"--bus={self.loader.busnum}",
                f"--device={self.loader.devnum}",
                mf.get_remote_path(),
            ],
            print_on_silent_log=True
        )



"""
This module implements the communication protocol to switch the digital outputs
on the Deditec Relais8 board.

Supported modules:

- Relais8

Supported Functionality:

- Turn digital output on and off
"""

import logging

from binascii import unhexlify, hexlify

import usb.core
import usb.util


class Relais8:
    _padding = "FEFEFEFEFEFEFEFEFF"
    _post_padding = "000000"

    _configuration_sequence = [
        "3430FF",
        "3434FF",
        "34C0FF",
        "3400FF",
        "3402FF",
        "340CFF",
        "340EFF",
        "3404FF",
        "3406FF",
        "3408FF",
        "340AFF",
        "3410FF",
        "3412FF",
        "3414FF",
        "3416FF",
        "34F4FF",
        "34F5FF",
        "34F6FF",
        "34F7FF",
        "34F0FF",
        "34ECFF",
    ]

    def __init__(self, **args):
        self._dev = usb.core.find(**args)
        if self._dev is None:
            raise ValueError("Device not found")
        if self._dev.is_kernel_driver_active(0):
            self._dev.detach_kernel_driver(0)

        cfg = self._dev.get_active_configuration()
        intf = cfg[(0, 0)]
        self._dev.set_configuration()
        usb.util.claim_interface(self._dev, 0)
        self._dev.ctrl_transfer(64, 0, 0, 0, 0, 5000)
        self._dev.ctrl_transfer(64, 3, 20, 0, 0, 5000)
        self._dev.ctrl_transfer(64, 9, 2, 0, 0, 5000)
        self._ep_out = usb.util.find_descriptor(
            intf, custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT
        )
        self._ep_in = usb.util.find_descriptor(
            intf, custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_IN
        )
        self._sequence_number = 0
        self._logger = logging.getLogger("Device: ")
        self._run_configuration_sequence()

    def _run_configuration_sequence(self):
        for d in Relais8._configuration_sequence:
            self.write_config_request(d)
            self.read(255)

    def read(self, size):
        data = self._ep_in.read(size)
        self._logger.debug("Read data: %s", hexlify(data))
        return data

    def write_config_request(self, address):
        complete_request = unhexlify(Relais8._padding)
        complete_request += bytes([self._sequence_number])
        complete_request += unhexlify(address + Relais8._post_padding)
        self._logger.debug("Sending Request: %s", hexlify(complete_request))
        self.write(complete_request)

    def write(self, data):
        self._sequence_number = (self._sequence_number + 1) & 0xFF
        self._logger.debug("Writing data: %s", hexlify(data))
        self._ep_out.write(data)

    def set_all_outputs(self, status, reset=False):
        assert isinstance(status, int)
        if status == 0:
            reset = True
        if status > 255 or status < 0:
            return
        data = f"0123{0 if reset else 8}000000001{status:02X}00000000000000"
        self.write(unhexlify(Relais8._padding + data))
        self.read(255)

    def set_output(self, number, status):
        if status:
            data = f"238000000001{1 << number - 1:02X}00000000000000"
        else:
            data = f"23A000000001{1 << number - 1:02X}00000000000000"
        complete_request = unhexlify(Relais8._padding)
        complete_request += bytes([self._sequence_number])
        complete_request += unhexlify(data)
        self.write(complete_request)
        self.read(255)

    def get_output(self, number):
        data = "34000000000F"
        complete_request = unhexlify(Relais8._padding)
        complete_request += bytes([self._sequence_number])
        complete_request += unhexlify(data)
        self.write(complete_request)
        read_data = self.read(255).tobytes()
        index = read_data.find(b"\x1a")
        if read_data[index + 1] == self._sequence_number - 1:
            outputs = read_data[index + 2]
        else:
            raise IOError("Could not read outputs from device")
        return outputs & (1 << (number - 1)) > 0

    def __del__(self):
        usb.util.release_interface(self._dev, 0)


def handle_set(busnum, devnum, number, status):
    relais8 = Relais8(bus=busnum, address=devnum)
    relais8.set_output(number, status)


def handle_get(busnum, devnum, number):
    relais8 = Relais8(bus=busnum, address=devnum)
    return relais8.get_output(number)


methods = {
    "set": handle_set,
    "get": handle_get,
}



import re

import pytest

from labgrid.driver import ExecutionError


def test_memory_mbw(command):
    """Test memcopy bandwidth"""
    try:
        command.run_check("which mbw")
    except ExecutionError:
        pytest.skip("mbw missing")

    result = command.run_check("mbw -qt0 8M")
    result = result[-1].strip()

    pattern = r"AVG\s+.*Copy:\s+(?P<bw>\S+)\s+MiB/s"
    (bw,) = map(float, re.fullmatch(pattern, result).groups())
    assert bw > 40  # > 40 MiB/second


def test_memory_memtester_short(command):
    """Test RAM for errors"""
    try:
        command.run_check("which memtester")
    except ExecutionError:
        pytest.skip("memtester missing")

    result = command.run_check("memtester 128k 1 | tail -n 1")
    result = result[-1].strip()

    assert result == "Done."



"""Tested with Aten PE6216.

HTTP API is defined by in the aten PDU PE6216 specification:
https://assets.aten.com/product/manual/Restful-API-Guide-for-PDU_2022-11-18.pdf
"""

import re
import requests

from ..exception import ExecutionError

PORT = 80

MIN_OUTLET_INDEX = 1
MAX_OUTLET_INDEX = 16
HEADERS = {"Content-Type": "application/x-www-form-urlencoded"}


def power_set(host, port, index, value):
    index = int(index)
    assert MIN_OUTLET_INDEX <= index <= MAX_OUTLET_INDEX
    value = "on" if value else "off"

    response = requests.post(
        f"http://{host}:{port}/api/outlet/relay",
        headers=HEADERS,
        data={
            "usr": "administrator",
            "pwd": "password",
            "index": index,
            "method": value,
        },
    )
    response.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert MIN_OUTLET_INDEX <= index <= MAX_OUTLET_INDEX

    response = requests.get(
        f"http://{host}:{port}/api/outlet/relay",
        headers=HEADERS,
        params={
            "usr": "administrator",
            "pwd": "password",
            "index": index,
        },
    )
    response.raise_for_status()

    m = re.search(r"<\d+>(?P<state>(ON|OFF|PENDING))<.*", response.text)
    if m is None:
        raise ExecutionError("PE6216: could not match response")
    return m.group("state") == "ON"



from labgrid.resource.flashrom import Flashrom, NetworkFlashrom
from labgrid.driver.flashromdriver import FlashromDriver


def test_flashrom_network_create(target):
    r = NetworkFlashrom(target, name=None, programmer="linux_spi:dev=/dev/spidev0.1,spispeed=30000", host="localhost")
    assert isinstance(r, NetworkFlashrom)
    d = FlashromDriver(target, name=None)
    assert isinstance(d, FlashromDriver)


def test_flashrom_driver_create(target):
    r = Flashrom(target, name=None, programmer="linux_spi:dev=/dev/spidev0.1,spispeed=30000")
    assert isinstance(r, Flashrom)

    d = FlashromDriver(target, name=None)
    assert isinstance(d, FlashromDriver)



import pytest

from labgrid.driver import SerialDriver
from labgrid.exceptions import NoSupplierFoundError


class TestSerialDriver:
    def test_instanziation_fail_missing_port(self, target):
        with pytest.raises(NoSupplierFoundError):
            SerialDriver(target, "serial")

    def test_instanziation(self, target, serial_port, mocker):
        serial_mock = mocker.patch("serial.Serial")
        s = SerialDriver(target, "serial")
        assert isinstance(s, SerialDriver)
        assert target.drivers[0] == s

    def test_write(self, target, serial_port, mocker):
        serial_mock = mocker.patch("serial.Serial")
        s = SerialDriver(target, "serial")
        target.activate(s)
        s.write(b"testdata")
        serial_mock.return_value.open.assert_called_once_with()
        serial_mock.return_value.write.assert_called_once_with(b"testdata")

    @pytest.mark.parametrize(
        "param",
        [
            [1, 1, None, 1],  # old test case
            [3, 2, None, 3],
            [3, 2, 5, 3],
            [3, 2, 1, 1],
            [1, 2, 1, 1],
        ],
    )
    # param = [size, in_waiting, max_size, out]
    def test_read(self, target, serial_port, mocker, param):
        serial_mock = mocker.patch("serial.Serial")
        serial_mock.return_value.in_waiting = param[1]
        s = SerialDriver(target, "serial")
        target.activate(s)
        s.read(size=param[0], max_size=param[2])
        serial_mock.return_value.open.assert_called_once_with()
        # assert 'read' called once with correct return:
        serial_mock.return_value.read.assert_called_once_with(param[3])

    def test_close(self, target, serial_port, mocker):
        serial_mock = mocker.patch("serial.Serial")
        s = SerialDriver(target, "serial")
        target.activate(s)
        s.close()
        serial_mock.return_value.open.assert_called_once_with()
        serial_mock.return_value.close.assert_called_once_with()

    def test_deactivate(self, target, serial_port, mocker):
        serial_mock = mocker.patch("serial.Serial")
        s = SerialDriver(target, "serial")
        target.activate(s)
        target.deactivate(s)
        serial_mock.return_value.open.assert_called_once_with()
        serial_mock.return_value.close.assert_called_once_with()

    def test_rfc2711_instanziation(self, target, serial_rfc2711_port, mocker):
        serial_mock = mocker.patch("serial.rfc2217.Serial")
        s = SerialDriver(target, "serial")
        assert isinstance(s, SerialDriver)
        assert target.drivers[0] == s
        serial_mock.assert_called_once_with()

    def test_raw_instanziation(self, target, serial_raw_port, mocker):
        serial_mock = mocker.patch("serial.serial_for_url")
        s = SerialDriver(target, "serial")
        assert isinstance(s, SerialDriver)
        assert target.drivers[0] == s
        serial_mock.assert_called_once_with("socket://", do_not_open=True)



import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..step import step
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class ManualSwitchDriver(Driver, DigitalOutputProtocol):
    description = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.status = False

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        if self.description is not None:
            description = self.description
        else:
            description = self.name

        self.target.interact(
            f"Set {description} for target {self.target.name} to {'ON' if status else 'OFF'} and press enter"  # pylint: disable=line-too-long
        )
        self.status = status

    @Driver.check_active
    @step(result=True)
    def get(self):
        return self.status



import pytest


def test_watchdog(command):
    stdout, stderr, returncode = command.run("wd 1")
    if returncode == 127:
        pytest.skip("wd command not available")
    assert returncode == 0
    assert not stderr
    assert not stdout

    command._await_prompt()

    stdout = command.run_check("echo ${global.system.reset}")
    assert len(stdout) == 1
    assert stdout[0] == "WDG"
